###############################################################################
# Coscine Python SDK
# Copyright (c) 2020-2025 RWTH Aachen University
# Licensed under the terms of the MIT License
# For more information on Coscine visit https://www.coscine.de/.
###############################################################################

"""
Provides functions and classes around the handling of metadata.
"""

from __future__ import annotations
from typing import TYPE_CHECKING, TypeAlias
from datetime import date, datetime, time, timedelta
from threading import Lock
from decimal import Decimal
import random
import logging
from string import ascii_letters
from requests.compat import urlparse
from tabulate import tabulate
from isodate import parse_duration, parse_datetime, parse_date, parse_time
import rdflib
import rdflib.term
import pyshacl
from coscine.exceptions import NotFoundError
if TYPE_CHECKING:
    from coscine.client import ApiClient
    from coscine.resource import Resource


logger = logging.getLogger(__name__)


# Type Alias according to PEP 613 for supported metadata value types
FormType: TypeAlias = (
    bool | date | datetime | Decimal | int | float | str | time | timedelta
)

# XML Schema datatype to Python native type lookup table
XSD_TYPES: dict[str, type] = {
    "any": str,
    "anyURI": str,
    "anyType": str,
    "byte": int,
    "int": int,
    "integer": int,
    "long": int,
    "unsignedShort": int,
    "unsignedByte": int,
    "negativeInteger": int,
    "nonNegativeInteger": int,
    "nonPositiveInteger": int,
    "positiveInteger": int,
    "short": int,
    "unsignedLong": int,
    "unsignedInt": int,
    "double": float,
    "float": float,
    "decimal": Decimal,
    "boolean": bool,
    "date": date,
    "dateTime": datetime,
    "duration": timedelta,
    "gDay": datetime,
    "gMonth": datetime,
    "gMonthDay": datetime,
    "gYear": datetime,
    "gYearMonth": datetime,
    "time": time,
    "ENTITIES": str,
    "ENTITY": str,
    "ID": str,
    "IDREF": str,
    "IDREFS": str,
    "language": str,
    "Name": str,
    "NCName": str,
    "NMTOKEN": str,
    "NMTOKENS": str,
    "normalizedString": str,
    "QName": str,
    "string": str,
    "token": str
}


def xsd_to_python(xmltype: str) -> type:
    """
    Converts an XMLSchema XSD datatype string to a native
    Python datatype class instance.
    """
    if xmltype.startswith("http://www.w3.org/2001/XMLSchema#"):
        xmltype = urlparse(xmltype)[-1]
        if xmltype not in XSD_TYPES:
            raise ValueError(
                "Failed to convert XMLSchema XSD datatype "
                f"to native Python type: Unsupported type '{xmltype}'!"
            )
        return XSD_TYPES[xmltype]
    return str


class Instance:
    """
    A (vocabulary) instance is an entry inside of a vocabulary.
    It maps from a human-readable name to a unique uniform
    resource identifier.
    """

    _data: dict

    @property
    def graph_uri(self) -> str:
        """
        The uniform resource identifier of the graph.
        If entered in a web browser, it should yield the definition
        of the graph.

        Example:
        >>> http://www.dfg.de/dfg_profil/gremien/fachkollegien/faecher/
        """
        return self._data.get("graphUri") or ""

    @property
    def instance_uri(self) -> str:
        """
        The uniform resource identifier of the instance.
        If entered in a web browser, it should yield the definition
        of the instance.

        Example:
        >>> http://www.dfg.de/dfg_profil/gremien/fachkollegien/liste/
        >>> index.jsp?id=112#112-03
        -> The item with id 112 within the graph
        """
        return self._data.get("instanceUri") or ""

    @property
    def type_uri(self) -> str:
        """
        Identifies the type of instance.

        Example:
        >>> http://www.dfg.de/dfg_profil/gremien/fachkollegien/liste/
        >>> index.jsp?id=112#112-03
        -> Commonly the same as instance_uri
        """
        return self._data.get("typeUri") or ""

    @property
    def subclass_of(self) -> str:
        """
        Identifies the subclass of the instance.

        Example:
        >>> http://www.dfg.de/dfg_profil/gremien/fachkollegien/liste/
        >>> index.jsp?id=112
        -> subclass of the item
        """
        return self._data.get("subClassOfUri") or ""

    @property
    def name(self) -> str:
        """
        The display name of the instance.
        """
        return self._data.get("displayName") or ""

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.name


class Vocabulary:
    """
    The Vocabulary contains all instances of a class and provides
    an interface to easily check whether a term is contained in
    the set of instances and to query the respective instance.
    """

    _data: list[Instance]

    def __init__(self, data: list[Instance]) -> None:
        self._data = data

    def __str__(self) -> str:
        return "\n".join(self.keys())

    def __contains__(self, key: str) -> bool:
        return key in self.keys()

    def __getitem__(self, key: str) -> str:
        for entry in self._data:
            if entry.name == key:
                return entry.instance_uri
        raise KeyError(f"Key {key} is not contained in vocabulary!")

    def __iter__(self):
        for key in self.keys():
            yield key

    def graph(self) -> rdflib.Graph:
        """
        Returns the vocabulary as an rdflib knowledge graph.
        """
        graph = rdflib.Graph()
        for entry in self._data:
            graph.add((
                rdflib.URIRef(entry.instance_uri),
                rdflib.RDF.type,
                rdflib.URIRef(entry.type_uri)
            ))
            graph.add((
                rdflib.URIRef(entry.instance_uri),
                rdflib.RDFS.label,
                rdflib.Literal(entry.name)
            ))
            if entry.subclass_of:
                graph.add((
                    rdflib.URIRef(entry.instance_uri),
                    rdflib.RDFS.subClassOf,
                    rdflib.URIRef(entry.subclass_of)
                ))
        return graph

    def keys(self) -> list[str]:
        """
        Returns the list of keys that are contained inside
        of the vocabulary. This equals the set of names of the
        class instances.
        """
        return [entry.name for entry in self._data]

    def resolve(self, value: str) -> FormType:
        """
        This method takes a value and returns its corresponding key.
        It can be considered the reverse of Vocabulary[key] -> value,
        namely Vocabulary[value] -> key but that cannot be expressed
        in Python, hence this method.
        """
        for entry in self._data:
            if entry.instance_uri == value:
                return entry.name
        raise KeyError(f"Value {value} is not contained in vocabulary!")


class FormField:
    """
    A FormField represents a MetadataField that has been specified
    in an application profile. The FormField has numerous properties
    which restrict the range of values that can be assigned to
    a metadata field. It is thus very important for the validation
    of metadata and ensures the consistency of metadata.
    """

    client: ApiClient

    _data: dict
    _values: list[FormType | MetadataForm]
    _vocabulary: Vocabulary | None

    @property
    def path(self) -> str:
        """
        The path of the FormField, acting as a unique identifier.
        """
        return self._data.get("path") or ""

    @property
    def name(self) -> str:
        """
        The human-readable name of the field, as displayed in
        the Coscine web interface.
        """
        return self._data.get("name") or ""

    @property
    def order(self) -> int:
        """
        The order of appearance of the field. The metadata fields are
        often displayed in a list in some sort of user interface. This
        property simply states at which position the field should appear.
        """
        return int(self._data.get("order") or 1)

    @property
    def class_uri(self) -> str:
        """
        In case the field is controlled by a vocabulary, the class_uri
        specifies the link to the instances of the vocabulary. These
        can then be fetched via ApiClient.instances(class_uri)
        """
        return self._data.get("class") or ""

    @property
    def min_count(self) -> int:
        """
        The minimum count of values that the field must receive.
        If the count is greater than 0, the field is a required one,
        as it will always need a value.
        """
        return self._data.get("minCount") or 0

    @property
    def max_count(self) -> int:
        """
        The maximum amount of values that can be given to the field.
        """
        return self._data.get("maxCount") or 128

    @property
    def min_length(self) -> int:
        """
        Specifies the minumum required length of the value.
        For values of type string this would equal the minimum
        string length.
        """
        return self._data.get("minLength") or 0

    @property
    def max_length(self) -> int:
        """
        Specifies the maximum permissible length of the value.
        For values of type string this would equal the maximum
        string length.
        """
        return self._data.get("maxLength") or 4096

    @property
    def literals(self) -> list[rdflib.Literal]:
        """
        The field as rdflib.Literal ready for use with rdflib.
        The literal has the appropriate datatype as specified in
        the SHACL application profile. This should be used as Coscine
        is very strict with its verification: There is apparently
        a difference between xsd:int and xsd:integer, I kid you not!
        """
        # See also: https://rdflib.readthedocs.io/en/stable/rdf_terms.html
        return [
            rdflib.Literal(value, datatype=self.xsd_type)
            for value in self.values
        ]

    @property
    def identifiers(self) -> list[rdflib.Literal] | list[rdflib.URIRef]:
        """
        The list of values as rdflib identifiers.
        """
        if self.has_vocabulary:
            return [rdflib.URIRef(serial) for serial in self.serial]
        return self.literals

    @property
    def node(self) -> str:
        """
        The node property of the metadata field, if present.
        """
        return self._data.get("node") or ""

    @property
    def xsd_type(self) -> str | None:
        """
        The string representation of the xsd:datatype.
        For example: http://www.w3.org/2001/XMLSchema#int
        """
        return self._data.get("datatype") or None

    @property
    def datatype(self) -> type:
        """
        Restricts the datatype of values that can be assigned
        to the field.
        """
        if self.is_inherited:
            return MetadataForm
        return xsd_to_python(self._data.get("datatype") or "")

    @property
    def vocabulary(self) -> Vocabulary:
        """
        In the case that the field has a value for the class_uri property,
        it is controlled by a vocabulary.
        """
        if not self.has_vocabulary:
            raise NotFoundError(
                f"Field {self.name} is not controlled by a vocabulary!"
            )
        if not self._vocabulary:
            self._vocabulary = self.client.vocabulary(self.class_uri)
        return self._vocabulary

    @property
    def selection(self) -> list[str]:
        """
        Some fields have a predefined selection of values that the user
        can choose from. In that case other values are not permitted.
        """
        return self._data.get("selection", "").split("~,~") or []

    @property
    def language(self) -> str:
        """
        The language setting of the field. This influences the field
        name and the values of fields controlled by a vocabulary or
        selection.
        """
        return self._data.get("language") or "en"

    @property
    def has_vocabulary(self) -> bool:
        """
        Evaluates to True if the field values are controlled by a vocabulary.
        """
        return bool(self._data["class"])

    @property
    def has_selection(self) -> bool:
        """
        Evaluates to True if the field values are controlled
        by a predefined selection of values.
        """
        return bool(self._data["selection"])

    @property
    def is_controlled(self) -> bool:
        """
        Evaluates to True if the field is either controlled by
        a vocabulary or a selection.
        """
        return self.has_vocabulary or self.has_selection

    @property
    def is_required(self) -> bool:
        """
        Evaluates to True if the field must be assigned a value before
        it can be sent to Coscine alongside the other metadata.
        """
        return self.min_count > 0

    @property
    def is_inherited(self) -> bool:
        """
        Evaluates to True if the field refers to another MetadataForm.
        """
        return bool(self.node)

    @property
    def serial(self) -> list[str]:
        """
        Serializes the metadata value to Coscine format. That means
        that for vocabulary controlled fields, the human-readable
        value is translated to the machine-readable unique identifier.
        This property can also be set with the metadata value received
        by the Coscine API, which is already in machine-readable format
        and will be translated to human-readable internally.
        """
        return self.serialize()

    @serial.setter
    def serial(self, value: str) -> None:
        self.values = [self.deserialize(value)]

    @property
    def values(self) -> list[FormType | MetadataForm]:
        """
        This is the value of the metadata field in human-readable
        form. For the machine-readable form that is sent to Coscine
        use the property FormValue.serial!
        Setting a value can only be done by using the appropriate datatype.
        If the FormField.max_count is greater than 1, you may assign a list
        of values to the field.
        """
        return self._values

    @values.setter
    def values(
        self,
        value: FormType | list[FormType | MetadataForm] | MetadataForm
    ) -> None:
        if isinstance(value, (list, tuple)):
            self._values = []
            for item in value:
                self.append(item)
        else:
            self.validate(value)
            self._values = [value]

    def _merge_default(
        self,
        root_node: rdflib.term.Node,
        graph: rdflib.Graph
    ) -> None:
        for value in self.identifiers:
            graph.add((root_node, rdflib.URIRef(self.path), value))

    def _merge_inherited(
        self,
        root_node: rdflib.term.Node,
        graph: rdflib.Graph
    ) -> None:
        for form in self.values:
            assert isinstance(form, MetadataForm)
            container = rdflib.term.BNode()
            graph.add((root_node, rdflib.URIRef(self.path), container))
            graph.add((container, rdflib.RDF.type, rdflib.URIRef(self.node)))
            for field in form.fields():
                field.merge(container, graph)

    def merge(self, root_node: rdflib.term.Node, graph: rdflib.Graph) -> None:
        """
        Merges the field as a value to root_node into an RDFlib graph
        """
        if not bool(self.values):
            return
        if self.is_inherited:
            self._merge_inherited(root_node, graph)
        else:
            self._merge_default(root_node, graph)

    def strlist(self) -> list[str]:
        """
        """
        entries = []
        if not self.invisible:
            if self.is_inherited:
                controlled = "Inherited"
                #values = self.node
                values = "\n".join([str(v) for v in self.values])
            else:
                controlled = "Controlled" if self.is_controlled else "Default"
                values = "\n".join([str(v) for v in self.values])
            entries = [
                self.is_required,
                controlled,
                self.datatype.__name__,
                f"{self.min_count} - {self.max_count}",
                self.name,
                values
            ]
        return entries

    def validate(self, value: FormType | MetadataForm) -> None:
        """
        Validates whether the value matches the specification of the
        FormField. Does not return anything but instead raises all
        sorts of exceptions.
        """
        if not isinstance(value, self.datatype):
            raise TypeError(
                f"While setting value for field {self.name}: "
                f"Expected type {self.datatype} but got {type(value)}!"
            )
        if self.is_controlled:
            assert isinstance(value, str)
            if self.has_vocabulary and value not in self.vocabulary:
                raise ValueError(
                    f"The field '{self.name}' is controlled by a vocabulary. "
                    f"The value '{value}' that you have provided did not "
                    "match any of the entries in the vocabulary!"
                )
            if self.has_selection and value not in self.selection:
                raise ValueError(
                    f"The field '{self.name}' is controlled by a selection. "
                    f"The value '{value}' that you have provided did not "
                    "match any of the entries in the selection!"
                )

    def metadata_form(self) -> MetadataForm:
        """
        Returns the inherited child MetadataForm if the field is inherited.
        """
        if not self.is_inherited:
            raise TypeError(f"The field {self.name} is not inherited!")
        return MetadataForm(self.client.application_profile(self.node))

    @property
    def invisible(self) -> bool:
        """
        FormFields can be set to invisible in the Coscine resource metadata
        default value settings. Inivisble FormFields are not displayed.
        """
        return self._invisible

    @invisible.setter
    def invisible(self, value: bool) -> None:
        self._invisible = value

    def append(
        self,
        value: FormType | MetadataForm,
        serialized: bool = False
    ) -> None:
        """
        If the field accepts a list of values, one can use the append
        method to add another value to the end of that list.
        """
        if serialized:
            if not isinstance(value, str):
                raise TypeError("Serialized values must be strings!")
            value = self.deserialize(value)
        self.validate(value)
        self._values.append(value)

    def serialize(self) -> list[str]:
        """
        Serializes the form field values into machine readable format.
        """
        if self.has_vocabulary:
            serialized_values = []
            for value in self.values:
                if not isinstance(value, str):
                    raise TypeError(f"Value '{value}' is not a string")
                serialized_values.append(self.vocabulary[value])
            return serialized_values
        return [str(value) for value in self.values]

    def deserialize(self, value: str) -> FormType:
        """
        Unmarshals the value and returns the pythonic representation.
        """
        if self.has_vocabulary:
            return self.vocabulary.resolve(value)
        if self.datatype == datetime:
            return parse_datetime(value)
        if self.datatype == date:
            return parse_date(value)
        if self.datatype == time:
            return parse_time(value)
        if self.datatype == timedelta:
            return parse_duration(value)
        if self.datatype == int:
            return int(value)
        if self.datatype == float:
            return float(value)
        if self.datatype == Decimal:
            return Decimal(value)
        if self.datatype == bool:
            return value.lower() == "true"
        return value

    def __init__(self, client: ApiClient, data: dict) -> None:
        self.client = client
        self._data = data
        self._invisible = False
        self._values = []
        self._vocabulary = None

    def clear(self) -> None:
        """
        Clears all values of all metadata fields.
        """
        self._values = []


class ApplicationProfileInfo:
    """
    Many different application profiles are available in Coscine.
    To be able to get information on a specific application profile or all
    application profiles, the ApplicationProfileInfo datatype is
    provided.
    """

    _data: dict

    @property
    def uri(self) -> str:
        """
        The uri of the application profile.
        """
        return self._data.get("uri") or ""

    @property
    def name(self) -> str:
        """
        The human-readable name of the application profile.
        """
        return self._data.get("displayName") or ""

    @property
    def description(self) -> str:
        """
        A description of the application profile.
        """
        return self._data.get("description") or ""

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.name


class ApplicationProfile(ApplicationProfileInfo):
    """
    An application profile defines how metadata can be specified.

    Parameters
    ----------
    client
        A Coscine Python SDK ApiClient for access to settings and requests.
    data
        ApplicationProfileInfo data as received by Coscine.
    """

    _data: dict

    client: ApiClient
    graph: rdflib.Graph
    lock = Lock()

    @property
    def definition(self) -> str:
        """
        The actual application profile in text/turtle format.
        """
        return self._data["definition"].get("content") or ""

    def __init__(self, client: ApiClient, data: dict) -> None:
        super().__init__(data)
        self.client = client
        self.graph = rdflib.Graph()
        self.graph.bind("sh", "http://www.w3.org/ns/shacl#")
        self.graph.bind("dcterms", "http://purl.org/dc/terms/")
        self.graph.parse(data=self.definition, format="ttl")
        self._determine_target_class()
        self._resolve_imports()

    def __str__(self) -> str:
        return str(self.graph.serialize(format="ttl"))

    def _determine_target_class(self) -> None:
        """
        Figures out the target class of the application profile.
        If not target class is present, the application profile URI is used
        as a fallback.
        """
        results = self.query(
            r"SELECT ?targetClass WHERE { ?_ sh:targetClass ?targetClass . }"
        )
        if len(results) != 1 or len(results[0]) != 1:
            self._target_class = self.uri
        else:
            self._target_class = results[0][0]

    def _resolve_imports(self) -> None:
        """
        Recursively resolves owl:imports statements
        """
        for row in self.query(
            "SELECT ?url WHERE { ?target sh:node ?url . }",
            initBindings={"target": rdflib.URIRef(self.uri)}
        ):
            profile = self.client.application_profile(str(row[0]))
            self.graph += profile.graph

        # Previous implementation:
        #for row in self.query("SELECT ?url WHERE { ?_ owl:imports ?url . }"):
        #    profile = self.client.application_profile(str(row[0]))
        #    self.graph += profile.graph

    @property
    def target_class(self) -> str:
        """
        Returns the target class of the application profile.
        If no target class is present, the application profile URI is used
        as a fallback.
        """
        return self._target_class

    def query(self, query: str, **kwargs) -> list:
        """
        Performs a SPARQL query on the application profile and
        returns the results as a list of rows, with each row
        containing as many columns as selected in the SPARQL query.

        Warnings
        ---------
        Note that rdflib SPARQL queries are NOT thread-safe! Under the
        hood pyparsing is invoked, which leads to a lot of trouble if
        used in a multithreaded context. To avoid any problems the
        Coscine Python SDK employs a lock on this function - only one
        thread can use it at any given time.
        TODO: Open pull request at rdflib and make rdflib itself thread-safe.

        Parameters
        ----------
        query
            A SPARQL query string.
        **kwargs
            Any number of keyword arguments to pass onto rdflib.query()
        """
        with self.lock:
            results: rdflib.query.Result = self.graph.query(query, **kwargs)
            items: list[list] = []
            for row in results:
                assert isinstance(row, rdflib.query.ResultRow)
                values = [
                    column.toPython() if column else None
                    for column in row
                ]
                items.append(values)
            return items

    def fields(self) -> list[FormField]:
        """
        Returns the list of metadata fields with their properties as specified
        in the application profile.
        """
        fields = []
        for result in self.query(
            "SELECT ?path ?name ?order ?class ?minCount ?maxCount\n"
            "?minLength ?maxLength ?datatype\n"
            "(GROUP_CONCAT(?in; SEPARATOR=\"~,~\") as ?ins)\n"
            "(lang(?name) as ?lang)\n"
            "?node\n"
            "WHERE {\n"
            "    ?_ sh:path ?path ;\n"
            "       sh:name ?name .\n"
            "    OPTIONAL { ?_ sh:order ?order . } .\n"
            "    OPTIONAL { ?_ sh:class ?class . } .\n"
            "    OPTIONAL { ?_ sh:minCount ?minCount . } .\n"
            "    OPTIONAL { ?_ sh:maxCount ?maxCount . } .\n"
            "    OPTIONAL { ?_ sh:minLength ?minLength . } .\n"
            "    OPTIONAL { ?_ sh:maxLength ?maxLength . } .\n"
            "    OPTIONAL { ?_ sh:datatype ?datatype . } .\n"
            "    OPTIONAL { ?_ sh:in/rdf:rest*/rdf:first ?in . } .\n"
            "    OPTIONAL { ?_ sh:node ?node . } . \n"
            "}\n"
            "GROUP BY ?name\n"
            "ORDER BY ASC(?order)\n"
        ):
            if result[10] == self.client.language:
                data: dict = {
                    "path": result[0],
                    "name": result[1],
                    "order": result[2],
                    "class": result[3],
                    "minCount": result[4],
                    "maxCount": result[5],
                    "minLength": result[6],
                    "maxLength": result[7],
                    "datatype": result[8],
                    "selection": result[9],
                    "language": result[10],
                    "node": result[11]
                }
                fields.append(FormField(self.client, data))
        return fields


class FileMetadata:
    """
    The existing metadata to a file as returned by the Coscine API.
    This metadata is by default in machine-readable format and not
    human-readable.
    """

    _data: dict

    @property
    def path(self) -> str:
        """
        Path/Identifier of the metadata field.
        """
        return self._data.get("path") or ""

    @property
    def type(self) -> str:
        """
        Datatype of the value as a string.
        """
        return self._data.get("type") or ""

    @property
    def version(self) -> str:
        """
        Current metadata version string. The version is a Unix timestamp.
        """
        return self._data.get("version") or ""

    @property
    def versions(self) -> list[str]:
        """
        List of all metadata version strings. Versions are unix timestamps.
        """
        versions = self._data.get("availableVersions")
        return list(versions) if versions else []

    @property
    def created(self) -> datetime:
        """
        The timestamp when the metadata was assigned.
        """
        return datetime.fromtimestamp(float(self.version))

    @property
    def definition(self) -> str:
        """
        The actual metadata in rdf turtle format.
        """
        turtle = self._data["definition"].get("content") or ""
        return turtle

    @property
    def is_latest(self) -> bool:
        """
        Returns True if the current metadata is the newest metadata
        for the file.
        """
        return self.version == max(self.versions)

    def graph(self) -> rdflib.Graph:
        """
        The metadata parsed as rdflib graph.
        """
        return rdflib.Graph().parse(data=self.definition, format="ttl")

    def fixed_graph(self, resource: Resource) -> rdflib.Graph:
        """
        Patches the file metadata knowledge graph to include the
        file path as its root subject.
        """
        graph = rdflib.Graph().parse(data=self.definition, format="ttl")
        for s, p, o in iter(graph.triples((None, None, None))):
            base: str = (
                f"https://purl.org/coscine/resources/{resource.id}/"
                f"{self.path}/@type=metadata&version={self.version}"
            )
            graph.remove((s, p, o))
            graph.add((rdflib.URIRef(base), p, o))
        return graph

    def items(self) -> list[dict[str, str]]:
        """
        Returns the list of metadata values in the format:
        >>> [{
        >>>     "path": "...",
        >>>     "value": "...",
        >>>     "datatype": "..."
        >>> }]
        """
        results: rdflib.query.Result = self.graph().query(
            "SELECT ?root ?property (str(?value) as ?value) "
            "(datatype(?value) as ?type)\n"
            "WHERE {\n"
            "    ?root ?property ?value .\n"
            "    FILTER( ?property != rdf:type )\n"
            "}\n"
        )
        items = []
        for row in results:
            assert isinstance(row, rdflib.query.ResultRow)
            root = str(row[0])
            path = str(row[1])
            value = str(row[2])
            datatype = str(row[3])
            item = {
                "root": root,
                "path": path,
                "value": value,
                "datatype": datatype
            }
            items.append(item)
        return items

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.definition


class MetadataForm:
    """
    The metadata form makes the meatadata fields that have been
    defined in an application profile accessible to users.

    Parameters
    ----------
    resource
        Coscine resource instance
    fixed_values
        If set to true, the fixed values set in the resource are applied
        when creating the application profile.
        If set to false, they are ignored and an empty metadata form
        is returned.
    """

    _fields: list[FormField]
    _fixed_values: dict
    application_profile: ApplicationProfile

    def __init__(
        self,
        application_profile: ApplicationProfile,
        fixed_values: dict | None = None
    ) -> None:
        self.application_profile = application_profile
        self._fields = self.application_profile.fields()
        if fixed_values:
            self._fixed_values = fixed_values
            self.defaults()

    def __str__(self) -> str:
        entries = []
        for key in self.keys():
            columns = self.field(key).strlist()
            if columns:
                entries.append(columns)
        headers: list[str] = [
            "Required", "Mode", "Type", "Range", "Field", "Value"
        ]
        return tabulate(entries, headers=headers, disable_numparse=True)

    def __setitem__(
        self,
        key: str, values: FormType | MetadataForm | list[FormType | MetadataForm]
    ) -> None:
        self.field(key).values = values

    def __getitem__(self, key: str) -> list[FormType | MetadataForm]:
        return self.field(key).values

    def __delitem__(self, key: str) -> None:
        self.field(key).clear()

    def __contains__(self, key: str) -> bool:
        return key in self.keys()

    def __iter__(self):
        for key in self.keys():
            yield key

    def defaults(self) -> None:
        """
        Parses the fixed and default value settings of a resource.
        This also includes visibility settings for metadata fields.
        """
        self.clear()
        for field_path, values in self._fixed_values.items():
            try:
                field = self.path(field_path)
            except KeyError:
                logger.warning(f"Fixed value path '{field_path}' is invalid!")
                continue
            if "https://purl.org/coscine/defaultValue" in values:
                field.values = [
                    field.deserialize(v["value"])
                    for v in values["https://purl.org/coscine/defaultValue"]
                    if v["type"] != "bnode"
                ]
            elif "https://purl.org/coscine/fixedValue" in values:
                field.values = [
                    field.deserialize(v["value"])
                    for v in values["https://purl.org/coscine/fixedValue"]
                    if v["type"] != "bnode"
                ]
            if "https://purl.org/coscine/invisible" in values:
                setting = values["https://purl.org/coscine/invisible"]
                field.invisible = bool(int(setting[0]["value"]))

    def clear(self) -> None:
        """
        Clears all values.
        """
        for field in self.fields():
            field.clear()

    def fields(self) -> list[FormField]:
        """
        The list of metadata fields that can be filled in as defined
        in the application profile.
        """
        return self._fields

    def field(self, key: str) -> FormField:
        """
        Looks up a metadata field via its name.
        """
        for field in self._fields:
            if field.name == key:
                return field
        raise KeyError(f"The field {key} is not part of the form!")

    def path(self, path: str) -> FormField:
        """
        Looks up a metadata field via its path.
        """
        for field in self._fields:
            if field.path == path:
                return field
        raise KeyError(f"The field path {path} is not part of the form!")

    def paths(self) -> list[str]:
        """
        Returns the list of paths of all metadata fields.
        """
        return [field.path for field in self._fields]

    def keys(self) -> list[str]:
        """
        Returns the list of names of all metadata fields.
        """
        return [field.name for field in self._fields]

    def values(self) -> list[list[FormType | MetadataForm]]:
        """
        Returns the list of values of all metadata fields.
        """
        return [field.values for field in self._fields]

    def items(self) -> list[tuple[str, list[FormType | MetadataForm]]]:
        """
        Returns key, value pairs for all metadata fields
        """
        return list(zip(self.keys(), self.values()))

    def graph(self) -> rdflib.Graph:
        """
        Returns the metadata as a knowledge graph.
        """
        root = rdflib.BNode()
        graph = rdflib.Graph()
        graph.add((
            root,
            rdflib.RDF.type,
            rdflib.URIRef(self.application_profile.target_class)
        ))
        for field in self._fields:
            field.merge(root, graph)
        return graph

    def validate(self) -> bool:
        """
        Validates the metadata against the resource application profile SHACL.
        """
        ontologies = rdflib.Graph()
        for field in self.fields():
            if field.has_vocabulary:
                ontologies += field.vocabulary.graph()
        graph = self.graph() + ontologies
        conforms, _, results_text = pyshacl.validate(
            graph,
            shacl_graph=self.application_profile.graph,
            ont_graph=ontologies,
            debug=False,
            inference="rdfs",
            abort_on_first=False,
            allow_infos=False,
            allow_warnings=False,
            meta_shacl=True,
            advanced=False,
            js=False
        )
        if not conforms:
            raise ValueError(results_text)
        return conforms

    def test(self) -> None:
        """
        Auto-fills the MetadataForm with a set of predefined values.
        Every field is filled in.
        """
        def generate_value(field: FormField):
            length = field.min_length if field.min_length > 1 else 8
            if length > field.max_length:
                length = field.max_length
            sample_data = {
                datetime: datetime.now(),
                date: datetime.now().date(),
                time: datetime.now().time(),
                timedelta: timedelta(2),
                int: random.randint(1, 16),
                float: random.random() * 123.0,
                Decimal: Decimal(random.randint(1, 42)),
                bool: True,
                str: "".join(random.choices(ascii_letters, k=length))
            }
            if field.has_vocabulary:
                length = len(field.vocabulary.keys()) - 1
                return field.vocabulary.keys()[random.randint(0, length)]
            if field.has_selection:
                length = len(field.selection) - 1
                return field.selection[random.randint(0, length)]
            if field.is_inherited:
                f = field.metadata_form()
                f.test()
                return f
            if field.datatype == "str" and field.min_length > 0:
                return "X" * field.min_length
            if field.datatype == "str" and field.max_length < 4096:
                return "X" * (field.max_length - 1)
            return sample_data[field.datatype]

        for field in self.fields():
            values = []
            if field.min_count > 1:
                for _ in range(field.min_count):
                    values.append(generate_value(field))
            else:
                values.append(generate_value(field))
            field.values = values

    def serialize(self, path: str) -> dict:
        """
        Prepares and validates metadata for sending to Coscine.
        Requires the file path of the file in Coscine as an argument.

        Parameters
        ----------
        path
            The path in Coscine to the FileObject that you would like
            to attach metadata to.
        """
        return {
            "path": path,
            "definition": {
                "content": self.graph().serialize(format="ttl"),
                "type": "text/turtle"
            }
        }

    def _parse_items(self, items: list[dict]) -> None:
        # Select the smallest lexicographic blank node guid
        active_root = min([i["root"] for i in items])
        for item in items:
            path: str = item["path"]
            value: str = item["value"]
            root: str = item["root"]
            if root == active_root:
                field = self.path(path)
                if field.is_inherited:
                    form = field.metadata_form()
                    subitems = list(filter(lambda e: e["root"] == value, items))
                    form._parse_items(subitems)
                    field.append(form)
                else:
                    field.append(value, True)

    def parse(self, data: FileMetadata) -> None:
        """
        Parses existing metadata that was received from Coscine.
        """
        self.clear()
        self._parse_items(data.items())



###############################################################################
# Coscine Python SDK
# Copyright (c) 2020-2025 RWTH Aachen University
# Licensed under the terms of the MIT License
# For more information on Coscine visit https://www.coscine.de/.
###############################################################################

"""
Provides the Coscine ApiClient.
"""

from __future__ import annotations
from typing import Iterator
from os import getenv, environ
import logging
from datetime import date, timedelta
import requests
import requests_cache
from requests.compat import quote as urlquote
from requests.utils import get_encoding_from_headers
from requests.adapters import HTTPAdapter
import urllib3
from urllib3.util import Retry
from coscine.__about__ import __version__
from coscine.exceptions import (
    AuthenticationError,
    CoscineException,
    NotFoundError,
    RequestRejected,
    TooManyResults,
)
from coscine.common import (
    ApiToken,
    Discipline,
    Language,
    License,
    MaintenanceNotice,
    Organization,
    SearchResult,
    User,
    Visibility,
)
from coscine.metadata import (
    ApplicationProfile,
    ApplicationProfileInfo,
    Instance,
    Vocabulary,
)
from coscine.project import Project, ProjectRole
from coscine.resource import ResourceType


logger = logging.getLogger(__name__)


class ApiResponse:
    """
    Models the response data object sent by the Coscine REST API upon
    a successful request.
    """

    response: requests.Response

    @property
    def data(self) -> dict:
        """
        The response data as a dict if it arrived in JSON format.
        """
        data = self.json
        return data["data"] if "data" in data else {}

    @property
    def json(self) -> dict:
        """
        The full response data as a dict if it arrived in JSON format.
        Includes response metadata.
        """
        try:
            return self.response.json()
        except requests.exceptions.JSONDecodeError:
            return {}

    @property
    def status_code(self) -> int:
        """
        The status code of the response as set by Coscine.
        """
        return self.json.get("statusCode") or 0

    @property
    def trace_id(self) -> str:
        """
        The Trace ID for Coscine internal error handling.
        """
        return self.json.get("traceId") or ""

    @property
    def is_paginated(self) -> bool:
        """
        Evaluates to True if the response is paginated, i.e. divided onto
        multiple pages.
        """
        return "pagination" in self.json

    @property
    def current_page(self) -> int:
        """
        The page number of the current page.
        """
        if self.is_paginated:
            return self.json["pagination"].get("currentPage") or 0
        raise AttributeError("The response is not paginated!")

    @property
    def total_pages(self) -> int:
        """
        The total number of pages for the specified PageSize.
        The PageSize is by default set to the maximum of 50.
        """
        if self.is_paginated:
            return self.json["pagination"].get("totalPages") or 0
        raise AttributeError("The response is not paginated!")

    @property
    def page_size(self) -> int:
        """
        The page size of the current response data.
        """
        if self.is_paginated:
            return self.json["pagination"].get("pageSize") or 0
        raise AttributeError("The response is not paginated!")

    @property
    def total_data_count(self) -> int:
        """
        The total amount of data items available.
        """
        if self.is_paginated:
            return self.json["pagination"].get("totalCount") or 0
        raise AttributeError("The response is not paginated!")

    @property
    def has_next(self) -> bool:
        """
        Evaluates to True if there are more pages available.
        """
        if self.is_paginated:
            return self.json["pagination"].get("hasNext") or False
        raise AttributeError("The response is not paginated!")

    @property
    def has_previous(self) -> bool:
        """
        Evaluates to True if there is at least one preceding page available.
        """
        if self.is_paginated:
            return self.json["pagination"].get("hasPrevious") or False
        raise AttributeError("The response is not paginated!")

    def pages(self) -> Iterator[ApiResponse]:
        """
        Returns all pages of the response. This may result in additional
        requests.
        """
        yield self
        if self.is_paginated:
            response = self
            request = self.request
            page = 1
            while response.has_next and page < self.client.max_pages:
                page += 1
                request.params["PageNumber"] = page
                response = self.client.send_request(request)
                yield response

    def __init__(
        self, client: ApiClient, request: requests.Request, response: requests.Response
    ) -> None:
        self.client = client
        self.request = request
        self.response = response


class ApiClient:
    """
    The ApiClient communicates with Coscine by sending requests to and
    receiving response data from Coscine.

    Parameters
    -----------
    token : str
        To be able to use the Coscine REST API one has to supply their
        Coscine API token. Every Coscine user can create their own set
        of API tokens for free in their Coscine user profile.
        For security reasons Coscine API Tokens are only valid for
        a certain amount of time after which they are deactivated.
    language : str
        Just like in the web interface of Coscine one can select
        a language preset for the Coscine API. This will localize
        all multi-language vocabularies and application profiles to the
        selected language. The language can later be switched on the fly.
    base_url : str
        Coscine is Open Source software and hosted on various domains.
        Via the base_url setting, the API user can specify which instance
        they would like to connect to. By default this is set to
        the Coscine instance of RWTH Aachen.
    verify_certificate : bool
        Whether to verify the SSL server certificate of the Coscine server.
        By default this is enabled as it provides some form of protection
        against spoofing, but on test instances or "fresh" installs
        certificates are often not used. To be able to use
        the Coscine Python SDK with such instances, the verify setting
        should be turned off.
    verbose : bool
        By disabling the verbose setting one can stop the Python SDK
        from printing to the command line interface (stdout). The stderr
        file handle is unaffected by this. This setting is particulary
        helpful in case you wish to disable the banner on initialization.
    enable_caching : bool
        Enabling caching allows the Python SDK to store some of
        the responses it gets from Coscine in a RequestCache. This cache
        is always active at runtime but may be saved and loaded to a file
        by enabling the caching setting. Entries in the cachefiles
        are valid for a certain amount of time until they are refreshed.
        With caching enabled the Python SDK is much faster.
    timeout : float (seconds)
        The timeout threshold for Coscine to respond to a request.
        If Coscine does not answer within the specified amount of
        seconds, an exception is raised. Note that setting a timeout
        is very important since otherwise your application may hang
        indefinitely if it never gets a response.
    use_native
        If enabled, up- and downloads are performed via the native providers.
        In the case of an s3 resource that equates to using boto3 behind
        the scenes. If set to False, the route via Coscine is taken, which
        is usually about 30% slower, less stable and has size and bandwidth
        limitations.
    """

    BANNER: str = (
        r"                      _              "
        "\n"
        r"                     (_)             "
        "\n"
        r"    ___ ___  ___  ___ _ _ __   ___   "
        "\n"
        r"   / __/ _ \/ __|/ __| | '_ \ / _ \  "
        "\n"
        r"  | (_| (_) \__ \ (__| | | | |  __/  "
        "\n"
        r"   \___\___/|___/\___|_|_| |_|\___|  "
        "\n"
        r" ___________________________________ "
        "\n"
        f"  Coscine Python SDK {__version__}   "
        "\n"
        r"  https://coscine.de/                "
        "\n"
    )

    base_url: str
    session: requests_cache.CachedSession
    verbose: bool
    verify: bool
    timeout: float

    _language: str

    @property
    def language(self) -> str:
        """
        The language setting of the ApiClient.
        This may be set to "en" for english or "de" for german.
        By default it is set to english but it can be changed
        on the fly even after the ApiClient has been instantiated.
        """
        return self._language

    @language.setter
    def language(self, value: str) -> None:
        language = value.lower()
        if language not in ("de", "en"):
            raise ValueError(
                f"Invalid language value: {value}! "
                "Acceptable values are: 'de' or 'en'."
            )
        self._language = language

    @property
    def version(self) -> str:
        """
        Coscine Python SDK version string. For example: "1.0.0"
        """
        return __version__

    def __init__(
        self,
        token: str,
        language: str = "en",
        base_url: str = "https://coscine.rwth-aachen.de",
        enable_caching: bool = True,
        verbose: bool = True,
        verify_certificate: bool = True,
        timeout: float = 60.0,
        retries: int = 3,
        use_native: bool = True,
        max_pages: int = 65535,
    ) -> None:
        self.base_url = base_url
        self.language = language
        self.timeout = timeout
        self.session = requests_cache.CachedSession(
            cache_name="coscine",
            allowable_methods=["GET"],
            backend="filesystem",
            serializer="json",
            use_cache_dir=True,
            urls_expire_after={
                "*/api/v2/application-profiles/profiles*": timedelta(days=30),
                "*/api/v2/disciplines*": timedelta(days=30),
                "*/api/v2/languages*": timedelta(days=30),
                "*/api/v2/licenses*": timedelta(days=30),
                "*/api/v2/organizations*": timedelta(days=30),
                "*/api/v2/resource-types/types*": timedelta(days=30),
                "*/api/v2/roles*": timedelta(days=30),
                "*/api/v2/titles*": timedelta(days=30),
                "*/api/v2/visibilities*": timedelta(days=30),
                "*/api/v2/vocabularies*": timedelta(days=30),
                "*": requests_cache.DO_NOT_CACHE,
            },
        )
        self.enable_caching(enable_caching)
        self.session.headers.update(
            {
                "Authorization": f"Bearer {token}",
                "User-Agent": f"Coscine Python SDK {self.version}",
            }
        )
        retrycfg = Retry(
            total=retries,
            backoff_factor=1.0,
            status_forcelist=[502, 503, 504],
            allowed_methods={"GET"},
        )
        self.session.mount("https://", HTTPAdapter(max_retries=retrycfg))
        self.verbose = verbose
        self.verify = verify_certificate
        self.native = use_native
        self.max_pages = max_pages
        if use_native and not getenv("AWS_REQUEST_CHECKSUM_CALCULATION"):
            environ["AWS_REQUEST_CHECKSUM_CALCULATION"] = "WHEN_REQUIRED"
        if not verify_certificate:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        if self.verbose:
            print(ApiClient.BANNER)
        try:
            self.self()
        except CoscineException:
            raise AuthenticationError("Invalid Coscine REST API Token!")

    def enable_caching(self, enable: bool) -> None:
        """
        Enables or disables request caching on demand.
        """
        self.session.settings.disabled = not enable

    def latest_version(self) -> str:
        """
        Retrieves the version string of the latest version of this
        package hosted on PyPi. Useful for checking whether the currently
        used version is outdated and if an update should be performed.

        Examples
        --------
        >>> if client.version != client.latest_version():
        >>>     print("Module outdated.")
        >>>     print("Run 'py -m pip install --upgrade coscine'.")
        """
        uri = "https://pypi.org/pypi/coscine/json"
        data = requests.get(uri, timeout=60.0).json()
        version = data["info"]["version"]
        return version

    def uri(self, *args) -> str:
        """
        Constructs a URI for requests to the Coscine REST API.
        This method creates URLs relative to the ApiClient.base_url
        and escapes URL arguments for compliance with the HTTP.

        Parameters
        -----------
        *args
            Any number of arguments that should be included in the URI.
            The arguments do not have to be of type string, but should
            be str() serializable.

        Examples
        --------
        >>> ApiClient.uri("application-profiles", "profiles", profile_uri)
        """
        suffix = "/".join((urlquote(str(arg), safe="") for arg in args))
        return f"{self.base_url}/coscine/api/v2/{suffix}"

    def send_request(
        self, request: requests.Request, stream: bool = False
    ) -> ApiResponse:
        """
        Sends a requests request to Coscine.

        Parameters
        ----------
        request
            The request that previously has been created
            with requests.Request().
        stream
            If set to True, the data transfer will be streamed,
            i.e. performed in chunks.
        """
        try:
            prepared_request = self.session.prepare_request(request)
            if prepared_request.body:
                encoding = get_encoding_from_headers(request.headers)
                if encoding:
                    assert isinstance(prepared_request.body, bytes)
                    body = prepared_request.body.decode(encoding)
                else:
                    body = "<binary data>"
                logger.debug(body)
            response = self.session.send(
                prepared_request,
                stream=stream,
                verify=self.verify,
                timeout=self.timeout,
            )
            encoding = get_encoding_from_headers(response.headers)
            if encoding:
                logger.debug(response.content.decode("utf-8"))
            response.raise_for_status()
            return ApiResponse(self, request, response)
        except requests.exceptions.RequestException as error:
            self.handle_request_exception(error)
            raise error

    def request(
        self, method: str, *args, stream: bool = False, **kwargs
    ) -> ApiResponse:
        """
        Sends a request to the Coscine REST API. This method is used
        internally. As a user of the ApiClient you should use the methods
        ApiClient.get(), ApiClient.post(), ApiClient.put(),
        ApiClient.delete(), ApiClient.options()
        instead of directly calling ApiClient.request().

        Parameters
        ----------
        method
            The HTTP method to use for the request:
            GET, PUT, POST, DELETE, OPTIONS, etc.
        *args
            Any number of arguments to forward to the requests.Request()
        stream
            If set to true, the response will be streamed. This means that
            the response will be split up and arrive in multiple chunks.
            When attempting to download files, the stream parameter must be
            set to True. Otherwise it should be left at False.
        *kwargs
            Any number of keyword arguments to forward to requests.Request()

        Raises
        ------
        See coscine.ApiClient.handle_request_exception
        """
        request = requests.Request(method, *args, **kwargs)
        return self.send_request(request, stream)

    @staticmethod
    def handle_request_exception(exception: Exception) -> None:
        """
        Raises
        ------
        ConnectionError
            If the request never reaches Coscine or we never get a response.
        coscine.exceptions.AuthenticationError
            If the Coscine API token is not valid.
        coscine.exceptions.RequestRejected
            If the request reached Coscine but was subsequently rejected
            by the Server.
        """
        if isinstance(exception, requests.exceptions.ConnectionError):
            raise ConnectionError(
                "Failed to connect to Coscine! Check your internet "
                "connection or whether Coscine is currently down."
            ) from exception
        if isinstance(exception, requests.exceptions.Timeout):
            raise ConnectionError(
                "The connection timed out. This may either be "
                "a ConnectionError or Coscine took a lot of time "
                "processing the request. You can increase the timeout "
                "threshold of the SDK in the client settings."
            ) from exception
        if isinstance(exception, requests.exceptions.RequestException):
            assert hasattr(exception, "response")
            assert isinstance(exception.response, requests.Response)
            if exception.response.status_code == 401:
                raise AuthenticationError(
                    "Invalid Coscine API token! The token was rejected "
                    "by Coscine. Check whether it is expired."
                ) from exception
            raise RequestRejected(
                "Coscine rejected the request sent by the Coscine Python SDK "
                "with the following error message: "
                f"{exception.response.content.decode('utf-8')}."
            ) from exception

    def get(self, *args, **kwargs) -> ApiResponse:
        """
        Sends a GET request to the Coscine REST API.
        For a list of exceptions that are raised by this method
        or a detailed list of parameters, refer to ApiClient.request().

        Parameters
        ----------
        Refer to ApiClient.request() for a list of parameters.

        Raises
        -------
        Refer to ApiClient.request() for a list of exceptions.

        Returns
        --------
        dict
            The "data": { ... } section of the JSON-response.
        """
        if "params" not in kwargs:
            kwargs["params"] = {}
        kwargs["params"]["PageSize"] = 50
        return self.request("GET", *args, **kwargs)

    def put(self, *args, **kwargs) -> ApiResponse:
        """
        Sends a PUT request to the Coscine REST API.

        Parameters
        ----------
        Refer to ApiClient.request() for a list of parameters.

        Raises
        -------
        Refer to ApiClient.request() for a list of exceptions.
        """
        return self.request("PUT", *args, **kwargs)

    def post(self, *args, **kwargs) -> ApiResponse:
        """
        Sends a POST request to the Coscine REST API.

        Parameters
        ----------
        Refer to ApiClient.request() for a list of parameters.

        Raises
        -------
        Refer to ApiClient.request() for a list of exceptions.
        """
        return self.request("POST", *args, **kwargs)

    def delete(self, *args, **kwargs) -> ApiResponse:
        """
        Sends a DELETE request to the Coscine REST API.

        Parameters
        ----------
        Refer to ApiClient.request() for a list of parameters.

        Raises
        -------
        Refer to ApiClient.request() for a list of exceptions.
        """
        return self.request("DELETE", *args, **kwargs)

    def options(self, *args, **kwargs) -> ApiResponse:
        """
        Sends an OPTIONS request to the Coscine REST API.

        Parameters
        ----------
        Refer to ApiClient.request() for a list of parameters.

        Raises
        -------
        Refer to ApiClient.request() for a list of exceptions.
        """
        return self.request("OPTIONS", *args, **kwargs)

    def projects(self, toplevel: bool = True) -> list[Project]:
        """
        Retrieves a list of all Coscine projects that the creator of
        the Coscine API token is currently a member of.

        Parameters
        -----------
        toplevel
            If set to True, only toplevel projects are retrieved.
            Set it to False to include all (sub-)projects in the results.
        """
        uri = self.uri("projects")
        response = self.get(uri, params={"TopLevel": toplevel})
        return [Project(self, item) for page in response.pages() for item in page.data]

    def _fetch_project_via_id(self, id: str) -> Project:
        uri = self.uri("projects", id)
        response = self.get(uri, params={"IncludeSubProjects": True})
        return Project(self, response.data)

    def project(
        self,
        key: str,
        attribute: property = Project.display_name,
        toplevel: bool = True,
    ) -> Project:
        """
        Returns a single Coscine Project via one of its properties.

        Parameters
        ----------
        key
            The value of the property to filter by.
        property
            The property/attribute of the project to filter by. Defaults to the shortend, 25 character display name of the project.
        toplevel
            If set to True, only toplevel projects are searched.
            Set it to False to include all (sub-)projects in the search.

        Raises
        ------
        coscine.exceptions.TooManyResults
            In case more than 1 project was found via the selected property.
        coscine.exceptions.NotFoundError
            In case the project could not be found via the selected property.
        """
        if attribute == Project.id:
            return self._fetch_project_via_id(key)
        results = list(
            filter(
                lambda project: project.match(attribute, key), self.projects(toplevel)
            )
        )
        if len(results) > 1:
            raise TooManyResults(
                "Found more than 1 project with a property matching "
                f"the key '{key}'. "
                "Certain project properties such as the name "
                "allow for duplicates among other projects. "
                "Use a different (unique) property to filter by!"
            )
        if len(results) == 0:
            raise NotFoundError(
                f"Failed to find a project via the key '{key}'! "
                "Maybe you are looking for a subproject and have "
                "set the toplevel argument to False? Also check whether "
                "you are filtering by the correct property."
            )
        # Required for access to subproject list
        return self._fetch_project_via_id(results[0].id)

    def create_project(
        self,
        name: str,
        display_name: str,
        description: str,
        start_date: date,
        end_date: date,
        principal_investigators: str,
        visibility: Visibility,
        disciplines: list[Discipline],
        responsible_organization: Organization,
        additional_organizations: list[Organization] | None = None,
        keywords: list[str] | None = None,
        grant_id: str = "",
        parent_id: str = "",
    ) -> Project:
        """
        Creates a new Coscine project.

        Parameters
        ----------
        name
            The project's name.
        display_name
            The project's display name (how it appears in the web interface).
        description
            The project description.
        start_date
            Date when the project starts.
        end_date
            Date when the project ends.
        principal_investigators
            The project PIs.
        visibility
            Project metadata visibility (relevant for search).
        disciplines
            List of associated scientific disciplines.
        responsible_organization
            The lead organization.
        additional_organizations
            List of organizations partaking in the project. The responsible
            organization may not re-appear here.
        keywords
            List of project keywords (relevant for search).
        grant_id
            The projects grant ID.
        parent_id
            Parent project ID if the project should be a subproject.
        """
        # Responsible organization and additional organizations are given
        # as a single list, with the differentiating aspect of an additional
        # responsible = True field for the responsible organization.
        organizations = []
        organizations.append(responsible_organization.serialize())
        organizations[0]["responsible"] = True
        if additional_organizations is None:
            additional_organizations = []
        for org in additional_organizations:
            data = org.serialize()
            data["responsible"] = False
            organizations.append(data)
        data: dict = {
            "name": name,
            "displayName": display_name,
            "description": description,
            "startDate": start_date.isoformat(),
            "endDate": end_date.isoformat(),
            "principleInvestigators": principal_investigators,
            "disciplines": [discipline.serialize() for discipline in disciplines],
            "organizations": organizations,
            "visibility": visibility.serialize(),
            "keywords": keywords if keywords else [],
            "grantId": grant_id,
        }
        if parent_id:
            data["parentId"] = parent_id
        uri = self.uri("projects")
        return Project(self, self.post(uri, json=data).data)

    def search(self, query: str, category: str | None = None) -> list[SearchResult]:
        """
        Sends a search request to Coscine and returns the results.

        Parameters
        -----------
        query
            The search query
        category
            The search can optionally be restricted to one of these
            categories: "metadata", "project" or "resource"
        """
        uri = self.uri("search")
        parameters = {"Query": query, "category": category}
        response = self.get(uri, params=parameters)
        return [SearchResult(item) for item in response.data]

    def maintenances(self) -> list[MaintenanceNotice]:
        """
        Retrieves the list of current active maintenance notices for Coscine.
        """
        uri = self.uri("maintenances")
        response = self.get(uri)
        return [MaintenanceNotice(item) for item in response.data]

    def visibilities(self) -> list[Visibility]:
        """
        Retrieves the list of visibility options available in Coscine.
        """
        uri = self.uri("visibilities")
        response = self.get(uri)
        return [Visibility(item) for item in response.data]

    def visibility(self, name: str) -> Visibility:
        """
        Returns the visibility that matches the name.
        Valid names are:
        * "Project Members"
        * "Public"
        """
        results = list(
            filter(lambda visibility: visibility.name == name, self.visibilities())
        )
        if len(results) > 1:
            raise TooManyResults(
                f"Found more than 1 visibility with the name '{name}'!"
            )
        if len(results) == 0:
            raise NotFoundError(f"Failed to find a visibility with the name '{name}'!")
        return results[0]

    def disciplines(self) -> list[Discipline]:
        """
        Retrieves the list of scientific disciplines available in Coscine.
        """
        uri = self.uri("disciplines")
        response = self.get(uri)
        return [Discipline(item) for page in response.pages() for item in page.data]

    def discipline(self, name: str) -> Discipline:
        """
        Returns the discipline that matches the name.
        Valid names are:
        * "Ancient Cultures 101"
        * "History 102"
        * "Fine Arts, Music, Theatre and Media Studies (103)"
        * "Linguistics 104"
        * "Library Studies 105"
        * "Social and Cultural Anthropology, Non-European Cultures, Jewish Studies and Religious Studies 106"
        * "Theology 107"
        * "Philosophy 108"
        * "Educational Research 109"
        * "Psychology 110"
        * "Social Sciences 111"
        * "Economics 112"
        * "Jurisprudence 113"
        * "Basic Biological and Medical Research 201"
        * "Plant Sciences 202"
        * "Zoology 203"
        * "Microbiology, Virology and Immunology 204"
        * "Medicine 205"
        * "Neurosciences 206"
        * "Agriculture, Forestry and Veterinary Medicine 207"
        * "Molecular Chemistry 301"
        * "Chemical Solid State and Surface Research 302"
        * "Physical and Theoretical Chemistry 303"
        * "Analytical Chemistry, Method Development (Chemistry) 304"
        * "Biological Chemistry and Food Chemistry 305"
        * "Polymer Research 306"
        * "Condensed Matter Physics 307"
        * "Optics, Quantum Optics and Physics of Atoms, Molecules and Plasmas 308"
        * "Particles, Nuclei and Fields 309"
        * "Statistical Physics, Soft Matter, Biological Physics, Nonlinear Dynamics 310"
        * "Astrophysics and Astronomy 311"
        * "Mathematics 312"
        * "Atmospheric Science, Oceanography and Climate Research 313"
        * "Geology and Palaeontology 314"
        * "Geophysics and Geodesy 315"
        * "Geochemistry, Mineralogy and Crystallography 316"
        * "Geography 317"
        * "Water Research 318"
        * "Production Technology 401"
        * "Mechanics and Constructive Mechanical Engineering 402"
        * "Process Engineering, Technical Chemistry 403"
        * "Heat Energy Technology, Thermal Machines, Fluid Mechanics 404"
        * "Materials Engineering 405"
        * "Materials Science 406"
        * "Systems Engineering 407"
        * "Electrical Engineering and Information Technology 408"
        * "Computer Science 409"
        * "Construction Engineering and Architecture 410"
        I had to copy these manually from a DFG PDF!
        The absolute state of digitization in germany...
        Expect some typos!
        """
        results = list(
            filter(lambda discipline: discipline.name == name, self.disciplines())
        )
        if len(results) > 1:
            raise TooManyResults(
                f"Found more than 1 discipline with the name '{name}'!"
            )
        if len(results) == 0:
            raise NotFoundError(f"Failed to find a discipline with the name '{name}'!")
        return results[0]

    def api_tokens(self) -> list[ApiToken]:
        """
        Retrieves the list of Coscine API tokens that have been created
        by the owner of the same API token that was used to initialize
        the Coscine Python SDK ApiClient.
        """
        uri = self.uri("self", "api-tokens")
        return [ApiToken(item) for page in self.get(uri).pages() for item in page.data]

    def self(self) -> User:
        """
        Returns the owner of the Coscine API token that was used to
        initialize the ApiClient.
        """
        return User(self.get(self.uri("self")).data)

    def users(self, query: str) -> list[User]:
        """
        Searches for users.
        """
        response = self.get(self.uri("users"), params={"SearchTerm": query})
        return [User(item) for page in response.pages() for item in page.data]

    def application_profile(self, profile_uri: str) -> ApplicationProfile:
        """
        Retrieves a specific application profile via its uri.

        Parameters
        ----------
        profile_uri
            The uri of the application profile,
            e.g. https://purl.org/coscine/ap/base/
            The trailing slash is important!
        """
        # Since we are parsing the ApplicationProfile right away as ttl
        # there is not point in fetching it in any other format or letting
        # the user decide the format.
        uri = self.uri("application-profiles", "profiles", profile_uri)
        response = self.get(uri, params={"format": "text/turtle"})
        return ApplicationProfile(self, response.data)

    def application_profiles(self) -> list[ApplicationProfileInfo]:
        """
        Retrieves the list of all application profiles that are currently
        available in Coscine.
        """
        uri = self.uri("application-profiles", "profiles")
        return [
            ApplicationProfileInfo(item)
            for page in self.get(uri).pages()
            for item in page.data
        ]

    def vocabulary(self, class_uri: str) -> Vocabulary:
        """
        Retrieves the vocabulary for the class.

        Parameters
        ----------
        class_uri
            The instance class uri, e.g. http://purl.org/dc/dcmitype
        """
        uri = self.uri("vocabularies", "instances")
        params = {"Class": class_uri, "Language": self.language}
        instances = [
            Instance(item)
            for page in self.get(uri, params=params).pages()
            for item in page.data
        ]
        return Vocabulary(instances)

    def languages(self) -> list[Language]:
        """
        Retrieves all language options available in Coscine.
        """
        uri = self.uri("languages")
        return [Language(item) for page in self.get(uri).pages() for item in page.data]

    def organization(self, ror_uri: str) -> Organization:
        """
        Looks up an organization based on its ror uri.
        Some commonly used ROR uris:
        * "https://ror.org/04xfq0f34" <- RWTH Aachen
        * "https://ror.org/04tqgg260" <- FH Aachen
        * "https://ror.org/04t3en479" <- KIT
        * "https://ror.org/02kkvpp62" <- TUM
        * "https://ror.org/05n911h24" <- TU Darm
        * "https://ror.org/01k97gp34" <- TU Do
        * "https://ror.org/02gm5zw39" <- UKA
        * "https://ror.org/00rcxh774" <- Uni Köln
        * "https://ror.org/04mz5ra38" <- Uni Due
        * "https://ror.org/00pd74e08" <- Uni Mün
        """
        uri = self.uri("organizations", ror_uri)
        return Organization(self.get(uri).data)

    def licenses(self) -> list[License]:
        """
        Retrieves a list of all licenses available in Coscine.
        """
        uri = self.uri("licenses")
        return [License(item) for page in self.get(uri).pages() for item in page.data]

    def license(self, name: str) -> License:
        """
        Returns the license that matches the name.
        * "Apache License 2.0"
        * "BSD 2-clause 'Simplified' License"
        * "BSD 3-clause 'New' or 'Revised' License"
        * "CC BY 4.0 (Attribution)"
        * "CC BY NC 4.0 (Attribution-NonCommercial)"
        * "CC BY NC ND 4.0 (Attribution-NonCommercial-NoDerivatives)"
        * "CC BY NC SA 4.0 (Attribution-NonCommercial-ShareAlike)"
        * "CC BY ND 4.0 (Attribution-NoDerivatives)"
        * "CC BY SA 4.0 (Attribution-ShareAlike)"
        * "Eclipse Public License 1.0"
        * "GNU Affero General Public License v3.0"
        * "GNU General Public License v2.0"
        * "GNU General Public License v3.0"
        * "GNU Lesser General Public License v2.1"
        * "GNU Lesser General Public License v3.0"
        * "MIT License"
        * "Mozilla Public License 2.0"
        * "The Unlicense"
        """
        results = list(filter(lambda license: license.name == name, self.licenses()))
        if len(results) > 1:
            raise TooManyResults(f"Found more than 1 licenses with the name '{name}'!")
        if len(results) == 0:
            raise NotFoundError(f"Failed to find a license with the name '{name}'!")
        return results[0]

    def roles(self) -> list[ProjectRole]:
        """
        Returns all roles that are available in Coscine,
        e.g. "Member", "Owner", ...
        """
        response = self.get(self.uri("roles"))
        return [ProjectRole(item) for item in response.data]

    def role(self, name: str) -> ProjectRole:
        """
        Returns the role that matches the name.
        """
        results = list(filter(lambda role: role.name == name, self.roles()))
        if len(results) > 1:
            raise TooManyResults(f"Found more than 1 role with the name '{name}'!")
        if len(results) == 0:
            raise NotFoundError(f"Failed to find a role with the name '{name}'!")
        return results[0]

    def resource_types(self) -> list[ResourceType]:
        """
        Retrieves a list of all resource types available in Coscine.
        """
        uri = self.uri("resource-types", "types")
        return [
            ResourceType(item) for page in self.get(uri).pages() for item in page.data
        ]

    def resource_type(self, name: str) -> ResourceType:
        """
        Returns the ResourceType that matches the name.
        Here name refers to the resource specificType,
        e.g. "rdsrwth" instead of the general type "rds".
        Mapping between specificType -> generalType:
        * "rdss3rwth" -> "rdss3"
        * "linked" -> "linked"
        * "rdss3wormrwth" -> "rdss3worm"
        * "rdsrwth" -> "rds"
        * "rdstudo" -> "rds"
        * "gitlab" -> "gitlab"
        * "rdss3nrw" -> "rdss3"
        * "rdss3ude" -> "rdss3"
        * "rdsude" -> "rds"
        * "rdss3tudo" -> "rdss3"
        * "rdsnrw" -> "rds"
        * "dsnrwweb" -> "dsnrwweb"
        * "dsNrwS3"-> "dsNrwS3"
        * "dsnrws3worm"-> "dsnrws3worm"
        """
        results = list(
            filter(lambda rtype: rtype.specific_type == name, self.resource_types())
        )
        if len(results) > 1:
            raise TooManyResults(
                f"Found more than 1 resource types with the name '{name}'!"
            )
        if len(results) == 0:
            raise NotFoundError(
                f"Failed to find a resource type with the name '{name}'!"
            )
        return results[0]

    def validate_pid(self, pid: str) -> bool:
        """
        Checks the given PID for validity.
        """
        try:
            if pid.startswith("http://hdl.handle.net/"):
                pid = pid.partition("http://hdl.handle.net/")[2]
            prefix, postfix = pid.split("/")
            uri = self.uri("pids", prefix, postfix)
            return self.get(uri).data.get("isValid") or False
        except (CoscineException, ValueError):
            return False



import logging
from os import getenv
import coscine

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("test")

def main(fp):
    token = getenv("COSCINE_API_TOKEN")
    if not token:
        raise ValueError(
            "No Coscine API Token specified! "
            "Create a new environment variable with "
            "the name COSCINE_API_TOKEN!"
        )
    client = coscine.ApiClient(token)
    application_profiles = client.application_profiles()
    count = len(application_profiles)
    valid = 0
    invalid = 0
    for index, ap in enumerate(application_profiles):
        print(f"AP {index}/{count} {ap.uri}", file=fp)
        try:
            form = coscine.MetadataForm(client.application_profile(ap.uri))
            form.test()
            print(form, file=fp)
            is_valid = form.validate()
            print(f"Valid: {is_valid}", file=fp)
            valid += 1
        except Exception as ex:
            print(f"Valid: {False}", file=fp)
            print(ex, file=fp)
            invalid += 1
    print(f"Valid: {valid}", file=fp)
    print(f"Invalid: {invalid}", file=fp)
    print(f"Total: {count}", file=fp)

with open("test-metadata-form-results.txt", "w", encoding="utf-8") as fp:
    main(fp)



###############################################################################
# Coscine Python SDK
# Copyright (c) 2020-2025 RWTH Aachen University
# Licensed under the terms of the MIT License
# For more information on Coscine visit https://www.coscine.de/.
###############################################################################

""" """

from __future__ import annotations
from typing import TYPE_CHECKING
from os import mkdir
from os.path import isdir
from posixpath import join as join_paths
from datetime import date
from textwrap import wrap
from isodate import parse_datetime
from tabulate import tabulate
from coscine.common import Discipline, License, Organization, User, Visibility
from coscine.resource import Resource, ResourceType, ResourceQuota
from coscine.metadata import ApplicationProfileInfo
from coscine.exceptions import NotFoundError, TooManyResults

if TYPE_CHECKING:
    from coscine.client import ApiClient


class ProjectQuota:
    """
    Projects have a set of storage space quotas. This class models
    the quota data returned by Coscine.
    """

    _data: dict

    @property
    def project_id(self) -> str:
        """
        The ID of the associated project.
        """
        return self._data.get("projectId") or ""

    @property
    def total_used(self) -> int:
        """
        The total used storage space in bytes.
        """
        return int(self._data["totalUsed"]["value"] * 1024**3)

    @property
    def total_reserved(self) -> int:
        """
        The total reserved storage space in bytes.
        """
        return int(self._data["totalReserved"]["value"] * 1024**3)

    @property
    def allocated(self) -> int:
        """
        The allocated storage space in bytes.
        """
        return int(self._data["allocated"]["value"] * 1024**3)

    @property
    def maximum(self) -> int:
        """
        The maximum available storage space in bytes.
        """
        value = self._data["maximum"]["value"] or 0
        return value * int(1024**3)

    @property
    def resource_type(self) -> ResourceType:
        """
        The associated resource type.
        """
        return ResourceType(self._data["resourceType"])

    @property
    def resource_quotas(self) -> list[ResourceQuota]:
        """
        The list of used resource quotas for the project.
        """
        return [ResourceQuota(data) for data in self._data["resourceQuotas"]]

    def __init__(self, data: dict) -> None:
        self._data = data


class ProjectRole:
    """
    Models roles that can be assumed by project members
    within a Coscine project.
    """

    _data: dict

    @property
    def id(self) -> str:
        """
        Unique and constant Coscine-internal identifier of the role.
        """
        return self._data.get("id") or ""

    @property
    def name(self) -> str:
        """
        Name of the role.
        """
        return self._data.get("displayName") or ""

    @property
    def description(self) -> str:
        """
        Description for the role.
        """
        return self._data.get("description") or ""

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.name


class ProjectMember:
    """
    This class models the members of a Coscine project.
    """

    project: Project
    _data: dict

    @property
    def id(self) -> str:
        """
        Unique Coscine-internal project member identifier.
        """
        return self._data.get("id") or ""

    @property
    def user(self) -> User:
        """
        The user in Coscine that represents the project member.
        """
        return User(self._data["user"])

    @property
    def role(self) -> ProjectRole:
        """
        The role of the member within the project.
        """
        return ProjectRole(self._data["role"])

    @role.setter
    def role(self, role: ProjectRole) -> None:
        uri = self.project.client.uri("projects", self.project.id, "members", self.id)
        self.project.client.put(uri, json={"roleId": role.id})

    def __init__(self, project: Project, data: dict) -> None:
        self.project = project
        self._data = data

    def __str__(self) -> str:
        return f"{self.user.display_name} as {self.role}"


class ProjectInvitation:
    """
    Models external user invitations via email in Coscine.
    """

    _data: dict

    @property
    def id(self) -> str:
        """
        Unique Coscine-internal identifier for the invitation.
        """
        return self._data.get("id") or ""

    @property
    def expires(self) -> date:
        """
        Timestamp of when the invitation expires.
        """
        value = self._data.get("expirationDate") or ""
        return parse_datetime(value)

    @property
    def email(self) -> str:
        """
        The email address of the invited user.
        """
        return self._data.get("email") or ""

    @property
    def issuer(self) -> User:
        """
        The user in Coscine who sent the invitation.
        """
        return User(self._data["issuer"])

    @property
    def project_id(self) -> str:
        """
        Project ID of the project the invitation applies to.
        """
        return self._data["project"]["id"]

    @property
    def role(self) -> ProjectRole:
        """
        Role assigned to the invited user.
        """
        return ProjectRole(self._data["role"])

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return f"{self.email} as {self.role.name}"


class Project:
    """
    Projects in Coscine contains resources.
    """

    client: ApiClient

    _data: dict

    @property
    def id(self) -> str:
        """
        Unique Coscine-internal project identifier.
        """
        return self._data.get("id") or ""

    @property
    def name(self) -> str:
        """
        The full project name as set in the project settings.
        """
        return self._data.get("name") or ""

    @name.setter
    def name(self, value: str) -> None:
        self._data["name"] = value

    @property
    def display_name(self) -> str:
        """
        The shortened project name as displayed in
        the Coscine web interface.
        """
        return self._data.get("displayName") or ""

    @display_name.setter
    def display_name(self, value: str) -> None:
        self._data["displayName"] = value

    @property
    def description(self) -> str:
        """
        The project description.
        """
        return self._data.get("description") or ""

    @description.setter
    def description(self, value: str) -> None:
        self._data["description"] = value

    @property
    def principal_investigators(self) -> str:
        """
        The project investigators.
        """
        return self._data.get("principleInvestigators") or ""

    @principal_investigators.setter
    def principal_investigators(self, value: str) -> None:
        self._data["principleInvestigators"] = value

    @property
    def start_date(self) -> date:
        """
        Start of project lifecycle timestamp.
        """
        value = self._data.get("startDate") or "1998-01-01T00:00:00Z"
        return parse_datetime(value)

    @start_date.setter
    def start_date(self, value: date) -> None:
        self._data["startDate"] = value

    @property
    def end_date(self) -> date:
        """
        End of project lifecycle timestamp.
        """
        value = self._data.get("endDate") or "1998-01-01T00:00:00Z"
        return parse_datetime(value)

    @end_date.setter
    def end_date(self, value: date) -> None:
        self._data["endDate"] = value

    @property
    def keywords(self) -> list[str]:
        """
        Project keywords for better discoverability.
        """
        return self._data.get("keywords") or []

    @keywords.setter
    def keywords(self, value: list[str]) -> None:
        self._data["keywords"] = value

    @property
    def grant_id(self) -> str:
        """
        Project grant id.
        """
        return self._data.get("grantId") or ""

    @grant_id.setter
    def grant_id(self, value: str) -> None:
        self._data["grantId"] = value

    @property
    def slug(self) -> str:
        """
        Project slug - usually a combination out of original
        project name and some arbitrary Coscine-internal
        data appended to it.
        """
        return self._data.get("slug") or ""

    @property
    def pid(self) -> str:
        """
        Project Persistent Identifier.
        """
        return self._data.get("pid") or ""

    @property
    def creator(self) -> str:
        """
        Project creator user ID.
        """
        creator = self._data.get("creator")
        if creator:
            return creator.get("id") or ""
        return ""

    @property
    def created(self) -> date:
        """
        Timestamp of when the project was created.
        If 1998-01-01T00:00:00Z is returned, then the created() value is erroneous
        or missing.
        """
        value = self._data.get("creationDate") or "1998-01-01T00:00:00Z"
        return parse_datetime(value)

    @property
    def organizations(self) -> list[Organization]:
        """
        Organizations participating in the project.
        """
        values = self._data.get("organizations", [])
        return [Organization(data) for data in values]

    @property
    def disciplines(self) -> list[Discipline]:
        """
        Scientific disciplines the project is involved with.
        """
        values = self._data.get("disciplines") or []
        return [Discipline(data) for data in values]

    @disciplines.setter
    def disciplines(self, value: list[Discipline]) -> None:
        self._data["disciplines"] = [discipline.serialize() for discipline in value]

    @property
    def visibility(self) -> Visibility:
        """
        Project visibility setting.
        """
        return Visibility(self._data["visibility"])

    @visibility.setter
    def visibility(self, value: Visibility) -> None:
        self._data["visibility"] = value.serialize()

    @property
    def url(self) -> str:
        """
        Project URL - makes the project accessible in the web browser.
        """
        return f"{self.client.base_url}/p/{self.slug}"

    def __init__(self, client: ApiClient, data: dict) -> None:
        self.client = client
        self._data = data

    def __str__(self) -> str:
        return tabulate(
            [
                ("ID", self.id),
                ("Name", self.name),
                ("Display Name", self.display_name),
                ("Description", "\n".join(wrap(self.description))),
                (
                    "Principal Investigators",
                    "\n".join(wrap(self.principal_investigators)),
                ),
                ("Disciplines", "\n".join([str(it) for it in self.disciplines])),
                ("Organizations", "\n".join([str(it) for it in self.organizations])),
                ("Start Date", self.start_date),
                ("End Date", self.end_date),
                ("Date created", self.created),
                ("Creator", self.creator),
                ("Grant ID", self.grant_id),
                ("PID", self.pid),
                ("Slug", self.slug),
                ("Keywords", ",".join(self.keywords)),
                ("Visibility", self.visibility),
            ],
            disable_numparse=True,
        )

    def match(self, attribute: property, key: str) -> bool:
        """
        Attempts to match the project via the given property
        and property value.
        Filterable properties:
        * Project.id
        * Project.pid
        * Project.name
        * Project.display_name
        * Project.url

        Returns
        -------
        True
            If its a match ♥
        False
            Otherwise :(
        """
        if (
            (attribute is Project.id and self.id == key)
            or (attribute is Project.pid and self.pid == key)
            or (attribute is Project.name and self.name == key)
            or (attribute is Project.url and self.url == key)
            or ((attribute is Project.display_name) and (self.display_name == key))
        ):
            return True
        return False

    def serialize(self) -> dict:
        """
        Marshals the project metadata into machine-readable format.
        """
        return {
            "name": self.name,
            "displayName": self.display_name,
            "description": self.description,
            "startDate": self.start_date.isoformat(),
            "endDate": self.end_date.isoformat(),
            "principleInvestigators": self.principal_investigators,
            "disciplines": [discipline.serialize() for discipline in self.disciplines],
            "organizations": [
                organization.serialize() for organization in self.organizations
            ],
            "visibility": self.visibility.serialize(),
            "keywords": self.keywords,
            "grantId": self.grant_id,
            "slug": self.slug,
            "pid": self.pid,
        }

    def delete(self) -> None:
        """
        Deletes the project on the Coscine servers.
        Be careful when using this function in your code, as users
        should be prevented from accidentially triggering it!
        Best to prompt the user before calling this function on whether
        they really wish to delete their project.
        """
        uri = self.client.uri("projects", self.id)
        self.client.delete(uri)

    def resources(self) -> list[Resource]:
        """
        Retrieves a list of all resources of the project.
        """
        uri = self.client.uri("projects", f"{self.id}", "resources")
        return [
            Resource(self, item)
            for page in self.client.get(uri).pages()
            for item in page.data
        ]

    # This is now required to access s3 options as they are no longer
    # sent with the resources() endpoint due to performance optimizations
    # in the Coscine API. Results in an additional request for a specific
    # resource.
    def _fetch_specific_resource_via_id(self, id: str) -> Resource:
        uri = self.client.uri("projects", f"{self.id}", "resources", id)
        return Resource(self, self.client.get(uri).data)

    def resource(
        self, key: str, attribute: property = Resource.display_name
    ) -> Resource:
        """
        Returns a single resource via one of its properties.
        The key can be specified to match any of the ResourceProperty items.
        """
        if attribute == Resource.id:
            return self._fetch_specific_resource_via_id(key)
        resources = self.resources()
        results = list(
            filter(lambda resource: resource.match(attribute, key), resources)
        )
        if len(results) > 1:
            raise TooManyResults(
                f"Found more than 1 resource matching the key '{key}'. "
                "Certain properties such as the name of a resource "
                "allow for duplicates among other resources."
            )
        if len(results) == 0:
            raise NotFoundError(f"Failed to find a resource via the key '{key}'! ")
        # Required for access to the resourcetypeoptions DTO
        return self._fetch_specific_resource_via_id(results[0].id)

    def download(self, path: str = "./") -> None:
        """
        Downloads the project to the local directory path.
        """
        path = join_paths(path, self.display_name, "")
        if not isdir(path):
            mkdir(path)
        for resource in self.resources():
            resource.download(path)

    def add_member(self, user: User, role: ProjectRole) -> None:
        """
        Adds the project member of another project to the current project.
        The owner of the Coscine API token must be a member
        of the other project.
        """
        uri = self.client.uri("projects", self.id, "members")
        self.client.post(uri, json={"roleId": role.id, "userId": user.id})

    def remove_member(self, member: ProjectMember) -> None:
        """
        Removes the member from the project. Does not invalidate the member
        object in Python - it is up to the API user to not use that variable
        again.
        """
        uri = self.client.uri("projects", self.id, "members", member.id)
        self.client.delete(uri)

    def invite(self, email: str, role: ProjectRole) -> None:
        """
        Invites an external user via their email address to
        the Coscine project.
        """
        uri = self.client.uri("projects", self.id, "invitations")
        self.client.post(uri, json={"roleId": role.id, "email": email})

    def quotas(self) -> list[ProjectQuota]:
        """
        Returns the project storage quotas.
        """
        uri = self.client.uri("projects", self.id, "quotas")
        response = self.client.get(uri)
        return [ProjectQuota(item) for item in response.data]

    def members(self) -> list[ProjectMember]:
        """
        Returns the list of all members of the current project.
        """
        uri = self.client.uri("projects", self.id, "members")
        return [
            ProjectMember(self, item)
            for page in self.client.get(uri).pages()
            for item in page.data
        ]

    def invitations(self) -> list[ProjectInvitation]:
        """
        Returns the list of all outstanding project invitations.
        """
        uri = self.client.uri("projects", self.id, "invitations")
        return [
            ProjectInvitation(item)
            for page in self.client.get(uri).pages()
            for item in page.data
        ]

    def update(self) -> None:
        """
        Updates a Coscine project's settings.
        To update certain properties just access the properties
        of the coscine.Project class directly and call Project.update()
        when done.
        """
        uri = self.client.uri("projects", self.id)
        self.client.put(uri, json=self.serialize())

    def subprojects(self) -> list[Project]:
        """ """
        if not "subProjects" in self._data:
            project = self.client.project(self.id, Project.id)
            self._data = project._data
        return [
            Project(self.client, data) for data in self._data.get("subProjects", [])
        ]

    def subproject(self, key: str, attribute: property = display_name) -> Project:
        """
        Returns a single Coscine Project via one of its properties.

        Parameters
        ----------
        key
            The value of the property to filter by.
        property
            The property/attribute of the project to filter by.
        toplevel
            If set to True, only toplevel projects are searched.
            Set it to False to include all (sub-)projects in the search.

        Raises
        ------
        coscine.exceptions.TooManyResults
            In case more than 1 project was found via the selected property.
        coscine.exceptions.NotFoundError
            In case the project could not be found via the selected property.
        """
        if attribute == Project.id:
            return self.client._fetch_project_via_id(key)
        results = list(
            filter(lambda project: project.match(attribute, key), self.subprojects())
        )
        if len(results) > 1:
            raise TooManyResults(
                "Found more than 1 project with a property matching "
                f"the key '{key}'. "
                "Certain project properties such as the name "
                "allow for duplicates among other projects. "
                "Use a different (unique) property to filter by!"
            )
        if len(results) == 0:
            raise NotFoundError(
                f"Failed to find a project via the key '{key}'! "
                "Maybe you are looking for a subproject and have "
                "set the toplevel argument to False? Also check whether "
                "you are filtering by the correct property."
            )
        # Required for access to subproject list
        return self.client._fetch_project_via_id(results[0].id)

    def create_subproject(
        self,
        name: str,
        display_name: str,
        description: str,
        start_date: date,
        end_date: date,
        principal_investigators: str,
        visibility: Visibility,
        disciplines: list[Discipline],
        responsible_organization: Organization,
        additional_organizations: list[Organization] | None = None,
        keywords: list[str] | None = None,
        grant_id: str = "",
    ) -> Project:
        """
        Creates a new Coscine project.

        Parameters
        ----------
        name
            The project's name.
        display_name
            The project's display name (how it appears in the web interface).
        description
            The project description.
        start_date
            Date when the project starts.
        end_date
            Date when the project ends.
        principal_investigators
            The project PIs.
        visibility
            Project metadata visibility (relevant for search).
        disciplines
            List of associated scientific disciplines.
        responsible_organization
            The lead organization.
        additional_organizations
            List of organizations partaking in the project. The responsible
            organization may not re-appear here.
        keywords
            List of project keywords (relevant for search).
        grant_id
            The projects grant ID.
        parent_id
            Parent project ID if the project should be a subproject.
        """
        return self.client.create_project(
            name,
            display_name,
            description,
            start_date,
            end_date,
            principal_investigators,
            visibility,
            disciplines,
            responsible_organization,
            additional_organizations,
            keywords,
            grant_id,
            self.id,
        )

    def create_resource(
        self,
        name: str,
        display_name: str,
        description: str,
        license: License,
        visibility: Visibility,
        disciplines: list[Discipline],
        resource_type: ResourceType,
        quota: int,
        application_profile: ApplicationProfileInfo,
        usage_rights: str = "",
        keywords: list[str] | None = None,
    ) -> Resource:
        """
        Creates a new Coscine resource within the project.

        Parameters
        ----------
        name
            The full name of the resource.
        display_name
            The shortened display name of the resource.
        description
            The description of the resource.
        license
            License for the resource contents.
        visibility
            Resource metadata visibility (relevant for search).
        disciplines
            Associated/Involved scientific disciplines.
        resource_type
            The Cosciner resource type.
        quota
            Resource quota in GB (irrelevant for linked data resources).
        application_profile
            The metadata application profile for the resource.
        notes
            Data usage notes
        keywords
            Keywords (relevant for search).

        """
        rds_options = {
            "quota": {"value": quota, "unit": "https://qudt.org/vocab/unit/GibiBYTE"}
        }
        options = {
            "linked": {"linkedResourceTypeOptions": {}},
            "gitlab": {
                "gitlabResourceTypeOptions": {},  # currently unsupported
            },
            "rds": {"rdsResourceTypeOptions": rds_options},
            "rdss3": {"rdsS3ResourceTypeOptions": rds_options},
            "rdss3worm": {"rdsS3WormResourceTypeOptions": rds_options},
            "dsnrwweb": {"dataStorageNrwWebResourceTypeOptions": rds_options},
            "dsnrws3worm": {"dataStorageNrwS3WormResourceTypeOptions": rds_options},
            "dsnrws3": {"dataStorageNrwS3ResourceTypeOptions": rds_options},
        }
        data: dict = {
            "name": name,
            "displayName": display_name,
            "description": description,
            "keywords": keywords if keywords else [],
            "license": license.serialize(),
            "visibility": visibility.serialize(),
            "disciplines": [discipline.serialize() for discipline in disciplines],
            "resourceTypeId": resource_type.id,
            "resourceTypeOptions": options[resource_type.general_type],
            "applicationProfile": {"uri": application_profile.uri},
            "usageRights": usage_rights,
        }
        uri = self.client.uri("projects", self.id, "resources")
        return Resource(self, self.client.post(uri, json=data).data)

    def update_quota(
        self, resource_type: ResourceType, allocated: int, maximum: int
    ) -> None:
        """
        Updates the quota settings of a project.
        This might require Admin privileges (not Owner) when
        influencing the maximum value. The allocated value
        should be modifiable by project owners.
        """
        uri = self.client.uri("projects", self.id, "quotas", resource_type.id)
        self.client.put(
            uri,
            json={
                "allocated": {
                    "value": allocated,
                    "unit": "https://qudt.org/vocab/unit/GibiBYTE",
                },
                "maximum": {
                    "value": maximum,
                    "unit": "https://qudt.org/vocab/unit/GibiBYTE",
                },
            },
        )



###############################################################################
# Coscine Python SDK
# Copyright (c) 2020-2025 RWTH Aachen University
# Licensed under the terms of the MIT License
# For more information on Coscine visit https://www.coscine.de/.
###############################################################################

"""
The Coscine Python SDK provides a high-level interface
to the Coscine REST API.
"""

import logging
from coscine.__about__ import *
from coscine.exceptions import *
from coscine.client import *
from coscine.common import *
from coscine.metadata import *
from coscine.project import *
from coscine.resource import *

# Set up logging to /dev/null like a library is supposed to.
# This ensures that if no logger was configured, all the logging
# calls made by this library do not yield any output.
logging.getLogger(__name__).addHandler(logging.NullHandler())



###############################################################################
# Coscine Python SDK
# Copyright (c) 2020-2025 RWTH Aachen University
# Licensed under the terms of the MIT License
# For more information on Coscine visit https://www.coscine.de/.
###############################################################################

"""
Provides an interface around resources in Coscine.
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Callable, BinaryIO
from datetime import date, datetime
from pathlib import Path
from io import BytesIO, IOBase
from os import mkdir
from os.path import basename, isdir, splitext
from posixpath import join as join_paths
import logging
from textwrap import wrap
from urllib.parse import urlparse, parse_qs
from isodate import parse_datetime
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from tabulate import tabulate
from tqdm import tqdm
import boto3
import rdflib
from coscine.common import Discipline, License, Visibility
from coscine.metadata import ApplicationProfile, MetadataForm, FileMetadata
from coscine.exceptions import NotFoundError, RequestRejected

if TYPE_CHECKING:
    from coscine.client import ApiClient
    from coscine.project import Project


logger = logging.getLogger(__name__)


class ResourceQuota:
    """
    Models the Coscine resource quota data.
    """

    _data: dict

    @property
    def resource_id(self) -> str:
        """
        The associated Coscine resource id.
        """
        return self._data["resource"].get("id") or ""

    @property
    def used_percentage(self) -> float:
        """
        The ratio of used up quota in relation to the available quota.
        """
        value = self._data.get("usedPercentage") or 0.00
        return float(value)

    @property
    def used(self) -> int:
        """
        The used quota in bytes.
        """
        return int(self._data["used"]["value"])

    @property
    def reserved(self) -> int:
        """
        The reserved quota for the resource.
        """
        return int(self._data["reserved"]["value"] * 1024**3)

    def __init__(self, data: dict) -> None:
        self._data = data

    def serialize(self) -> dict:
        return self._data


class ResourceTypeOptions:
    """
    Options and settings regarding the resource type.
    Mainly provides an interface to resource type specific attributes
    such as S3 access credentials for resources of type rds-s3.
    """

    _data: dict

    @property
    def bucket_name(self) -> str:
        """
        The S3 bucket name.
        """
        return self._data.get("bucketName") or ""

    @property
    def access_key_read(self) -> str:
        """
        The S3 access key for reading.
        """
        return self._data.get("accessKeyRead") or ""

    @property
    def secret_key_read(self) -> str:
        """
        The S3 secret key for reading.
        """
        return self._data.get("secretKeyRead") or ""

    @property
    def access_key_write(self) -> str:
        """
        The S3 access key for writing.
        """
        return self._data.get("accessKeyWrite") or ""

    @property
    def secret_key_write(self) -> str:
        """
        The S3 secret key for writing.
        """
        return self._data.get("secretKeyWrite") or ""

    @property
    def endpoint(self) -> str:
        """
        The S3 endpoint.
        """
        return self._data.get("endpoint") or ""

    @property
    def size(self) -> int:
        """
        The size setting of the resource type in GibiByte.
        """
        if "size" in self._data:
            return int(self._data["size"].get("value"))
        return 0

    @size.setter
    def size(self, value: int) -> None:
        self._data["size"] = {
            "value": value,
            "unit": "https://qudt.org/vocab/unit/GibiBYTE",
        }

    def __init__(self, data: dict | None = None) -> None:
        self._data = data if data else {}


class ResourceType:
    """
    Models the resource types available in Coscine.
    """

    _data: dict

    @property
    def options(self) -> ResourceTypeOptions:
        """
        The resource's resource type specific options.
        """
        try:
            options = self._data["options"]
            if self.general_type == "rdss3":
                options = options.get("rdsS3")
            elif self.general_type == "rdss3worm":
                options = options.get("rdsS3Worm")
            elif self.general_type == "rds":
                options = options.get("rds")
            elif self.general_type == "dsnrws3":
                options = options.get("dsNrwS3")
            elif self.general_type == "dsnrws3worm":
                options = options.get("dsNrwS3Worm")
            elif self.general_type == "dsnrwweb":
                options = options.get("dsNrwWeb")
            elif self.general_type == "gitlab":
                options = options.get("gitLab")
            elif self.general_type == "linked":
                options = options.get("linkedData")
            return ResourceTypeOptions(options)
        except KeyError:
            return ResourceTypeOptions()

    @property
    def id(self) -> str:
        """
        Coscine-internal resource type identifier.
        """
        return self._data.get("id") or ""

    @property
    def general_type(self) -> str:
        """
        General resource type, e.g. rdss3
        """
        general_type = self._data.get("generalType") or ""
        return general_type.lower()

    @property
    def specific_type(self) -> str:
        """
        Specific resource type, e.g. rdss3rwth
        """
        specific_type = self._data.get("specificType") or ""
        return specific_type.lower()

    @property
    def active(self) -> str:
        """
        Whether the resource type is enabled on the Coscine instance.
        """
        return self._data.get("status") or ""

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.specific_type

    def serialize(self) -> dict[str, dict]:
        """
        Serializes to resourceTypeOptions {},
        not type.
        """
        return self._data


class Resource:
    """
    Models a Coscine Resource object.
    """

    client: ApiClient
    project: Project
    _data: dict

    @property
    def id(self) -> str:
        """
        Unique Coscine-internal resource identifier.
        """
        return self._data.get("id") or ""

    @property
    def name(self) -> str:
        """
        Full resource name as displayed in the resource settings.
        """
        return self._data.get("name") or ""

    @name.setter
    def name(self, value: str) -> None:
        self._data["name"] = value

    @property
    def display_name(self) -> str:
        """
        Shortened resource name as displayed in the Coscine web interface.
        """
        return self._data.get("displayName") or ""

    @display_name.setter
    def display_name(self, value: str) -> None:
        self._data["displayName"] = value

    @property
    def description(self) -> str:
        """
        The resource description.
        """
        return self._data.get("description") or ""

    @description.setter
    def description(self, value: str) -> None:
        self._data["description"] = value

    @property
    def type(self) -> ResourceType:
        """
        The resource's resource type.
        """
        return ResourceType(self._data["type"])

    @property
    def pid(self) -> str:
        """
        The persistent identifier assigned to the resource.
        """
        return self._data.get("pid") or ""

    @property
    def url(self) -> str:
        """
        Project URL - makes the resource accessible in the web browser.
        """
        return f"{self.client.base_url}/p/{self.project.slug}/r/{self.id}/-/"

    @property
    def access_url(self) -> str:
        """Resource Access URL via PID"""
        return f"http://hdl.handle.net/{self.pid}"

    @property
    def keywords(self) -> list[str]:
        """
        List of keywords for better discoverability.
        """
        return self._data.get("keywords") or []

    @keywords.setter
    def keywords(self, value: list[str]) -> None:
        self._data["keywords"] = value

    @property
    def license(self) -> License | None:
        """
        The license used for the resource data.
        """
        value = self._data.get("license")
        if value:
            return License(value)
        return None

    @license.setter
    def license(self, value: License) -> None:
        self._data["license"] = value.serialize()

    @property
    def usage_rights(self) -> str:
        """
        The usage rights specified for the data inside the resource.
        """
        return self._data.get("usageRights") or ""

    @usage_rights.setter
    def usage_rights(self, value: str) -> None:
        self._data["usageRights"] = value

    @property
    def application_profile(self) -> ApplicationProfile:
        """
        The application profile of the resource.
        """
        uri = self._data["applicationProfile"]["uri"]
        return self.client.application_profile(uri)

    @property
    def disciplines(self) -> list[Discipline]:
        """
        The scientific disciplines set for the resource.
        """
        values = self._data.get("disciplines", [])
        return [Discipline(data) for data in values]

    @disciplines.setter
    def disciplines(self, value: list[Discipline]) -> None:
        self._data["disciplines"] = [discipline.serialize() for discipline in value]

    @property
    def visibility(self) -> Visibility:
        """
        The Coscine visibility setting for the resource.
        """
        return Visibility(self._data["visibility"])

    @visibility.setter
    def visibility(self, value: Visibility) -> None:
        self._data["visibility"] = value.serialize()

    @property
    def created(self) -> date:
        """
        Timestamp of when the resource was created.
        """
        value = self._data.get("dateCreated") or "1998-01-01T00:00:00Z"
        return parse_datetime(value)

    @property
    def creator(self) -> str:
        """
        The Coscine user id of the resource creator.
        """
        return self._data.get("creator") or ""

    @property
    def archived(self) -> bool:
        """
        Evaluates to True when the resource is set to archived.
        """
        return bool(self._data.get("archived"))

    @archived.setter
    def archived(self, value: bool) -> None:
        self._data["archived"] = value

    @property
    def quota(self) -> ResourceQuota:
        """
        The resources storage quota.
        """
        uri = self.client.uri(
            "projects", self.project.id, "resources", self.id, "quota"
        )
        return ResourceQuota(self.client.get(uri).data)

    @property
    def fixed_values(self) -> dict:
        """
        The resources default metadata values.
        """
        return self._data.get("fixedValues") or {}

    def __init__(self, project: Project, data: dict) -> None:
        self.client = project.client
        self.project = project
        self._data = data

    def __str__(self) -> str:
        return tabulate(
            [
                ("ID", self.id),
                ("Name", self.name),
                ("Display Name", self.display_name),
                ("Description", "\n".join(wrap(self.description))),
                ("Disciplines", "\n".join([str(i) for i in self.disciplines])),
                ("Date created", self.created),
                ("Creator", self.creator),
                ("PID", self.pid),
                ("Keywords", self.keywords),
                ("Visibility", self.visibility),
                ("Application Profile", self.application_profile.name),
                ("Usage rights", self.usage_rights),
                ("License", self.license),
                ("Archived", self.archived),
            ],
            disable_numparse=True,
        )

    def match(self, attribute: property, key: str) -> bool:
        """
        Attempts to match the resource via the given property
        and property value.
        Filterable properties:
        * Resource.id
        * Resource.pid
        * Resource.name
        * Resource.display_name
        * Resource.url

        Returns
        -------
        True
            If its a match ♥
        False
            Otherwise :(
        """
        if (
            (attribute is Resource.id and self.id == key)
            or (attribute is Resource.pid and self.pid == key)
            or (attribute is Resource.name and self.name == key)
            or (attribute is Resource.url and self.url == key)
            or ((attribute is Resource.display_name) and (self.display_name == key))
        ):
            return True
        return False

    def serialize(self) -> dict:
        """
        Serializes Coscine Resource metadata into machine-readable
        representation.
        """
        data = {
            "name": self.name,
            "displayName": self.display_name,
            "description": self.description,
            "id": self.id,
            "pid": self.pid,
            "keywords": self.keywords,
            "visibility": self.visibility.serialize(),
            "disciplines": [discipline.serialize() for discipline in self.disciplines],
            "applicationProfile": {"uri": self.application_profile.uri},
            "type": self.type.serialize(),
            "usageRights": self.usage_rights,
            "fixedValues": self.fixed_values,
            "archived": self.archived,
            "quota": self.quota.serialize(),
        }
        if self.license:
            data["license"] = self.license.serialize()
        return data

    def metadata_form(self, fixed_values: bool = True) -> MetadataForm:
        """
        Returns the resource metadata form.

        Parameters
        ----------
        fixed_values
            If set to true, the fixed values set in the resource are applied
            when creating the application profile.
            If set to false, they are ignored and an empty metadata form
            is returned.
        """
        return MetadataForm(
            self.application_profile, self.fixed_values if fixed_values else None
        )

    def update(self) -> None:
        """
        Change the values locally via setter properties
        """
        uri = self.client.uri("projects", self.project.id, "resources", self.id)
        self.client.put(uri, json=self.serialize())

    def delete(self) -> None:
        """
        Deletes the Coscine resource and along with it all files
        and metadata contained within it on the Coscine servers.
        Special care should be taken when using that method in code
        as to not accidentially trigger a delete on a whole resource.
        Therefore this method is best combined with additional input
        from the user e.g. by prompting them with the message
        "Do you really want to delete the resource? (Y/N)".
        """
        uri = self.client.uri("projects", self.project.id, "resources", self.id)
        self.client.delete(uri)

    def download(self, path: str = "./") -> None:
        """
        Downloads the resource to the local directory given by path.
        """
        path = join_paths(path, self.display_name, "")
        if not isdir(path):
            mkdir(path)
        for file in self.files(recursive=True):
            file.download(path, True)

    def mkdir(self, path: str, metadata: MetadataForm | None = None) -> None:
        """
        Creates a folder inside of a resource. Should work for all
        resource types.
        """
        if not path.endswith("/"):
            path += "/"
        self.upload(path, "", metadata)

    def upload(
        self,
        path: str,
        handle: BinaryIO | bytes | str,
        metadata: MetadataForm | dict | rdflib.Graph | None = None,
        progress: Callable[[int], None] | None = None,
        overwrite: bool = False,
    ) -> None:
        """
        Uploads a file-like object to a resource in Coscine.

        Parameters
        ----------
        path
            The path the file shall assume inside of the Coscine resource.
            Not the path on your local harddrive!
            The terms path, key and filename can be used interchangeably.
        handle
            A binary file handle that supports reading or
            a set of bytes or a string that can be utf-8 encoded.
        metadata
            Metadata for the file that matches the resource
            application profile.
        progress
            Optional callback function that gets occasionally called
            during the upload with the progress in bytes.
        """
        if isinstance(metadata, MetadataForm):
            metadata = metadata.serialize(path)
        elif isinstance(metadata, dict):
            raise NotImplementedError("Fill MetadataForm")
        elif isinstance(metadata, rdflib.Graph):
            metadata = {
                "path": path,
                "definition": {
                    "content": metadata.serialize(format="turtle"),
                    "type": "text/turtle",
                },
            }
        else:
            raise ValueError("unrecognized metadata format")
        if metadata is None and self.type.general_type != "rdss3":
            raise ValueError(
                "Resources other than S3 resources require metadata " "for uploads!"
            )
        if isinstance(handle, str):
            handle = handle.encode("utf-8")
        if isinstance(handle, bytes):
            handle = BytesIO(handle)
        if metadata is not None:
            self.update_metadata(metadata, overwrite)
        assert isinstance(handle, IOBase)
        if self.type.general_type == "rdss3" and self.client.native:
            self._upload_blob_s3(path, handle, progress)
        else:
            self._upload_blob(path, handle, progress)

    def post_metadata(self, metadata: dict) -> None:
        """
        Creates metadata for a file object for the first time.
        There shall be no metadata assigned to the file already - in that
        case use put_metadata()!
        """
        uri = self.client.uri(
            "projects", self.project.id, "resources", self.id, "trees", "metadata"
        )
        self.client.post(uri, json=metadata)

    def put_metadata(self, metadata: dict) -> None:
        """
        Updates existing metadata of a file object. If the file object
        does not yet have metadata, use post_metadata()!
        """
        uri = self.client.uri(
            "projects", self.project.id, "resources", self.id, "trees", "metadata"
        )
        self.client.put(uri, json=metadata)

    def update_metadata(self, metadata: dict, overwrite: bool = False) -> None:
        """
        Updates metadata of a file. In case no metadata has yet been
        assigned to the file, it will create new metadata. This method
        basically incorporates both post_metadata() and put_metadata()
        into one, choosing the appropriate method when applicable.
        This comes at the cost of possibly sending two requests, where
        one would have sufficed.
        """
        if overwrite:
            try:
                self.put_metadata(metadata)
            except RequestRejected:
                self.post_metadata(metadata)
        else:
            self.post_metadata(metadata)

    def _upload_blob(
        self,
        path: str,
        handle: BinaryIO,
        progress: Callable[[int], None] | None = None,
        use_put: bool = False,
    ):
        """
        Uploads a file-like object to a resource in Coscine.

        Parameters
        ----------
        path : str
            The path the file shall assume inside of the Coscine resource.
            Not the path on your local harddrive!
            The terms path, key and filename can be used interchangeably.
        handle : BinaryIO
            A binary file handle that supports reading.
        metadata : MetadataForm or dict
            Metadata for the file that matches the resource
            application profile.
        progress : Callable "def function(int)"
            Optional callback function that gets occasionally called
            during the upload with the progress in bytes.
        """
        uri = self.client.uri(
            "projects", self.project.id, "resources", self.id, "blobs", path
        )
        files = {"file": (path, handle, "application/octect-stream")}
        encoder = MultipartEncoder(fields=files)
        progress_bar = tqdm(
            smoothing=0.1,
            mininterval=0.5,
            desc=path,
            total=encoder.len,
            unit="B",
            unit_scale=True,
            ascii=True,
            disable=not self.client.verbose,
        )

        def progress_callback(mon):
            nonlocal progress_bar, progress
            progress_bar.update(mon.bytes_read - progress_bar.n)
            if progress:
                progress(mon.bytes_read)

        monitor = MultipartEncoderMonitor(encoder, progress_callback)
        headers = {"Content-Type": monitor.content_type}
        if use_put:
            self.client.put(uri, data=monitor, headers=headers)
        else:
            self.client.post(uri, data=monitor, headers=headers)

    def _upload_blob_s3(
        self, path: str, handle: BinaryIO, progress: Callable[[int], None] | None = None
    ) -> None:
        """
        Works only on rdss3 resources and should not be called
        on other resource types! Bypasses Coscine and uploads
        directly to the underlying s3 storage.
        """
        progress_bar = tqdm(
            smoothing=0.1,
            mininterval=0.5,
            desc=path,
            unit="B",
            unit_scale=True,
            ascii=True,
            disable=not self.client.verbose,
        )
        s3 = boto3.client(
            "s3",
            aws_access_key_id=self.type.options.access_key_write,
            aws_secret_access_key=self.type.options.secret_key_write,
            endpoint_url=self.type.options.endpoint,
        )

        bytes_read_abs = 0

        def progress_callback(bytes_read_inc):
            nonlocal progress_bar, progress, bytes_read_abs
            progress_bar.update(bytes_read_inc)
            bytes_read_abs += bytes_read_inc
            if progress:
                progress(bytes_read_abs)

        s3.upload_fileobj(
            handle, self.type.options.bucket_name, path, Callback=progress_callback
        )

    def _fetch_files_recursively(self, path: str = ""):
        more_folders: bool = True
        contents: list[FileObject] = []
        directories: list[FileObject] = []
        for fileobj in self._fetch_files(path):
            if more_folders:
                more_folders = False
            if fileobj.is_folder:
                directories.append(fileobj)
                if fileobj.path != "/":
                    yield from self._fetch_files(fileobj.path)
                more_folders = True
            else:
                yield fileobj

    def _fetch_files(self, path: str = ""):
        uri = self.client.uri(
            "projects", self.project.id, "resources", self.id, "trees", "files"
        )
        params = {"Path": path} if path else {}
        for page in self.client.get(uri, params=params).pages():
            for item in page.data:
                yield FileObject(self, item)

    def files(
        self, path: str = "", recursive: bool = False, with_metadata: bool = False
    ) -> list[FileObject]:
        """
        Retrieves the list of files that are contained in the resource.
        Via an additional single API call the metadata for those files
        can be fetched and made available in the returned files.

        Parameters
        ----------
        path
            You can limit the set of returned files to a path.
            The path may be the path to a single file in which
            case a list containing that single file will be returned.
            Or it may point to a "directory" in which case all
            the files contained in that "directory" are returned.
        recursive
            S3 resources may have folders inside them. Set the recursive
            parameter to True to also fetch all files contained
            in these folders.
        with_metadata
            If set to True the set of files are returned alongside
            with their metadata. This internally requires another
            API request which is considerably slower (1 to 2 seconds).
            However if you plan on manipulating each files metadata
            this is the way to go. Otherwise you would have to make
            an API call to fetch the metadata for each file which
            in case of large resources will prove to be very painful... :)
        """
        if recursive:
            files = self._fetch_files_recursively(path)
        else:
            files = self._fetch_files(path)
        if with_metadata:
            metadata = self.metadata(path)
            for file in files:
                file.assign_metadata(metadata)
        return files

    def file(self, path: str) -> FileObject:
        """
        Returns a single file of the resource via its unique path.
        """
        uri = self.client.uri(
            "projects", self.project.id, "resources", self.id, "trees", "files"
        )
        files = [
            FileObject(self, item)
            for item in self.client.get(uri, params={"Path": path}).data
        ]
        if len(files) > 1:
            # path argument acts as prefix, may yield more than 1 result
            # Only return exact matches!
            for f in files:
                if f.path == path:
                    return f
            raise NotFoundError
        if len(files) == 0:
            raise NotFoundError
        return files[0]

    def metadata(self, path: str = "") -> list[FileMetadata]:
        """
        Returns the full set of metadata for each file in the resource.
        """
        uri = self.client.uri(
            "projects", self.project.id, "resources", self.id, "trees", "metadata"
        )
        params = {"Path": path} if path else {}
        return [
            FileMetadata(item)
            for page in self.client.get(uri, params=params).pages()
            for item in page.data
        ]

    def graph(self) -> rdflib.Graph:
        """
        Returns a knowledge graph with the full set of file object metadata.
        """
        graph = self.application_profile.graph
        for metadata in self.metadata():
            if metadata.is_latest:
                subgraph = metadata.fixed_graph(self)
                graph += subgraph
        return graph

    def query(self, sparql: str) -> list[FileObject]:
        """
        Runs a SPARQL query on the underlying resource knowledge graph and
        returns the file objects whose metadata matches the query.
        IMPORTANT: The query must (!) include ?path as a variable/column.
        Otherwise it will get rejected and a ValueError is raised.

        Examples
        --------
        >>> resource.query("SELECT ?path ?p ?o { ?path ?p ?o . }")

        >>> project = client.project("Solaris")
        >>> resource = project.resource("Chest X-Ray CNN")
        >>> files = resource.query(
        >>>     "SELECT ?path WHERE { "
        >>>     "    ?path dcterms:creator ?creator . "
        >>>     "     FILTER(?creator != 'Dr. Akula') "
        >>>     "}"
        >>> )
        >>> for file in files:
        >>>     print(file.path)
        """
        results: rdflib.query.Result = self.graph().query(sparql)
        columns: list[str] = [x.toPython() for x in results.vars]
        if "?path" not in columns:
            raise ValueError("?path not present in sparql query string!")
        files: list[FileObject] = []
        filepaths: list[str] = [row.path.split("/")[6] for row in results]
        for file in self.files(recursive=True, with_metadata=True):
            if file.path in filepaths:
                files.append(file)
        return files

    def file_index(self) -> list[dict]:
        """
        Returns a file index with the following data:
        {
            file-path: {
                filename: "foo",
                filesize: 0,
                download: http://example.org/foo,
                expires: datetime
            }, ...
        }
        This index can easily be serialized to JSON format and made available
        publicly. Coscine currently prohibits external users from
        downloading files in a resource.
        To be able to publicly make data available one can
        instead publish this file index, which needs to be updated in
        regular intervals to ensure that the download urls do not expire.
        By hosting the json-serialized representation of this index on
        a free to use platform such as GitHub, one can access it via browser
        or via software and thus use Coscine as a storage provider in software
        and data publications regardless of whether people have access
        to Coscine.
        """
        return [
            {
                "path": file.path,
                "filename": file.name,
                "filesize": file.size,
                "download": file.download_url,
                "expires": file.download_expires,
            }
            for file in self.files(recursive=True)
            if not file.is_folder
        ]


class FileObject:
    """
    Models files or file-like objects in Coscine resources.
    """

    _data: dict
    _metadata: FileMetadata | None
    resource: Resource

    @property
    def path(self) -> str:
        """
        The path to the file. Usually equivalent to the filename, except
        when the file is a directory or contained within a directory.
        Which in the case of S3 resource may occur regularly.

        Examples
        ---------
        >>> chest_xray.png
        >>> pneumonia/lung_ap.png
        >>> pneumonia/
        """
        return self._data.get("path") or ""

    @property
    def type(self) -> str:
        """
        The type of the file object in the file tree.

        Examples
        ---------
        >>> Leaf
        >>> Tree
        """
        return self._data.get("type") or ""

    @property
    def filetype(self) -> str:
        """
        The file's filetype.

        Examples
        --------
        >>> ".png"
        >>> ".txt"
        >>> ""
        """
        return splitext(self.name)[1]

    @property
    def directory(self) -> str:
        """
        The directory the file object is located in, if it is in a folder.
        """
        return self._data.get("directory") or ""

    @property
    def name(self) -> str:
        """
        The filename of the file. Includes the file type extension.

        Examples
        --------
        >>> foo.txt
        >>> bar.png

        """
        return self._data.get("name") or ""

    @property
    def extension(self) -> str:
        """
        The file type extension.
        """
        return self._data.get("extension") or ""

    @property
    def size(self) -> int:
        """
        The size of the file contents in bytes.
        """
        value = self._data.get("size") or 0
        return int(value)

    @property
    def created(self) -> date:
        """
        Timestamp of when the file has been uploaded.
        """
        value = self._data.get("creationDate") or "1998-01-01T00:00:00Z"
        return parse_datetime(value)

    @property
    def modified(self) -> date:
        """
        Timestamp of when the file was recently modified.
        """
        value = self._data.get("changeDate") or "1998-01-01T00:00:00Z"
        return parse_datetime(value)

    @property
    def is_folder(self) -> bool:
        """
        Evaluates to True when the FileObject represents a folder and
        not an actual file.
        """
        return self.type == "Tree"

    @property
    def download_url(self) -> str:
        """
        The download URL for the file.
        """
        if "actions" in self._data:
            return self._data["actions"]["download"]["url"]
        return ""

    @property
    def download_expires(self) -> datetime:
        """
        The timestamp when the FileObject.download_url will expire.
        """
        if self.download_url:
            expires = parse_qs(urlparse(self.download_url).query)["Expires"][0]
        else:
            expires = "0"
        return datetime.fromtimestamp(float(expires))

    @property
    def client(self) -> ApiClient:
        """
        The Coscine ApiClient associated with the resource instance.
        """
        return self.resource.client

    def __init__(
        self, resource: Resource, data: dict, metadata: FileMetadata | None = None
    ) -> None:
        self.resource = resource
        self._data = data
        self._metadata = metadata

    def __str__(self) -> str:
        return self.path

    def delete(self) -> None:
        """
        Deletes the FileObject remote on the Coscine server.
        """
        uri = self.client.uri(
            "projects",
            self.resource.project.id,
            "resources",
            self.resource.id,
            "blobs",
            self.path,
        )
        self.client.delete(uri)

    def metadata(self, refresh: bool = True) -> FileMetadata | None:
        """
        Returns the metadata of the file. This might use a cached version
        of file metadata or make a request, if no cached version is available.
        """
        if (not self._metadata) and refresh:
            data = self.resource.metadata(path=self.path)
            self.assign_metadata(data)
        return self._metadata

    def metadata_form(self, refresh: bool = True) -> MetadataForm:
        """
        Returns the metadata of the file or an empty metadata form if
        no metadata has been attached to the file.
        """
        form = MetadataForm(self.resource.application_profile)
        metadata = self.metadata(refresh=refresh)
        if metadata:
            form.parse(metadata)
        return form

    def download(self, path: str = "./", recursive: bool = False) -> None:
        """
        Downloads the file to the computer.
        If path ends in a filename, the whole path is used.
        Otherwise the filename of the file is appended to the given path.
        If recursive is True, the full path of the file is used and appended
        to the path. But all folders on that path must have already been
        created then.
        """
        if not basename(path):
            path = join_paths(path, self.path if recursive else self.name)
        filepath = path
        Path(filepath).parent.mkdir(parents=True, exist_ok=True)
        if self.is_folder:
            return
        else:
            # Since Coscine allows directories and files to share the same name
            # we need to handle collisions on the client side. Most filesystems
            # such as FAT and NTFS do not allow files to have the same name
            # as a folder within the same directory.
            for i in range(100):
                if not isdir(filepath):
                    break
                filepath = f"{path} ({i})"
            with open(filepath, "wb") as fp:
                self.stream(fp)

    def stream(self, fp: BinaryIO) -> None:
        """
        Streams file contents.
        """
        if self.resource.type.general_type == "rdss3" and self.client.native:
            self.stream_s3(fp)
        else:
            self.stream_blob(fp)

    def stream_blob(self, fp: BinaryIO) -> None:
        """
        Streams file contents from the Coscine Blob API.
        """
        uri = self.client.uri(
            "projects",
            self.resource.project.id,
            "resources",
            self.resource.id,
            "blobs",
            self.path,
        )
        progress_bar = tqdm(
            smoothing=0.1,
            mininterval=0.5,
            desc=self.path,
            total=self.size,
            unit="B",
            unit_scale=True,
            ascii=True,
            disable=not self.client.verbose,
        )
        response = self.client.request("GET", uri, stream=True)
        for chunk in response.response.iter_content(chunk_size=4096):
            progress_bar.update(len(chunk))
            fp.write(chunk)

    def stream_s3(self, handle: BinaryIO) -> None:
        """
        Works only on rdss3 resources and should not be called
        on other resource types! Bypasses Coscine and uploads
        directly to the underlying s3 storage.
        """
        progress_bar = tqdm(
            smoothing=0.1,
            mininterval=0.5,
            desc=self.path,
            total=self.size,
            unit="B",
            unit_scale=True,
            ascii=True,
            disable=not self.client.verbose,
        )
        s3 = boto3.client(
            "s3",
            aws_access_key_id=self.resource.type.options.access_key_write,
            aws_secret_access_key=self.resource.type.options.secret_key_write,
            endpoint_url=self.resource.type.options.endpoint,
        )
        s3.download_fileobj(
            self.resource.type.options.bucket_name,
            self.path,
            handle,
            Callback=progress_bar.update,
        )

    def update(
        self,
        handle: BinaryIO | bytes | str,
        progress: Callable[[int], None] | None = None,
    ) -> None:
        """
        Uploads a file-like object to a resource in Coscine.

        Parameters
        ----------
        handle : BinaryIO | bytes | str
            A binary file handle that supports reading or
            bytes or str.
        progress : Callable "def function(int)"
            Optional callback function that gets occasionally called
            during the upload with the progress in bytes.
        """
        if isinstance(handle, str):
            handle = handle.encode("utf-8")
        if isinstance(handle, bytes):
            handle = BytesIO(handle)

        if self.type.general_type == "rdss3" and self.client.native:
            self.resource._upload_blob_s3(self.path, handle, progress)
        else:
            self.resource._upload_blob(self.path, handle, progress, use_put=True)

    def update_metadata(self, metadata: MetadataForm | dict | rdflib.Graph) -> None:
        """
        Updates the metadata for a file object or creates new metadata
        if there has not been metadata assigned yet.
        """
        self._metadata = None
        if isinstance(metadata, MetadataForm):
            metadata = metadata.serialize(self.path)
        elif isinstance(metadata, dict):
            raise NotImplementedError("Fill MetadataForm")
        elif isinstance(metadata, rdflib.Graph):
            metadata = {
                "path": self.path,
                "definition": {
                    "content": metadata.serialize(format="turtle"),
                    "type": "text/turtle",
                },
            }
        else:
            raise ValueError("unrecognized metadata format")
        self.resource.update_metadata(metadata)

    def post_metadata(self, metadata: MetadataForm | dict | rdflib.Graph) -> None:
        """
        Creates metadata for a file object for the first time.
        There shall be no metadata assigned to the file already - in that
        case use put_metadata()!
        """
        if isinstance(metadata, MetadataForm):
            metadata = metadata.serialize(self.path)
        elif isinstance(metadata, dict):
            raise NotImplementedError("Fill MetadataForm")
        elif isinstance(metadata, rdflib.Graph):
            metadata = {
                "path": self.path,
                "definition": {
                    "content": metadata.serialize(format="turtle"),
                    "type": "text/turtle",
                },
            }
        else:
            raise ValueError("unrecognized metadata format")
        self.resource.post_metadata(metadata)

    def put_metadata(self, metadata: MetadataForm | dict | rdflib.Graph) -> None:
        """
        Updates existing metadata of a file object. If the file object
        does not yet have metadata, use post_metadata()!
        """
        if isinstance(metadata, MetadataForm):
            metadata = metadata.serialize(self.path)
        elif isinstance(metadata, dict):
            raise NotImplementedError("Fill MetadataForm")
        elif isinstance(metadata, rdflib.Graph):
            metadata = {
                "path": self.path,
                "definition": {
                    "content": metadata.serialize(format="turtle"),
                    "type": "text/turtle",
                },
            }
        else:
            raise ValueError("unrecognized metadata format")
        self.resource.put_metadata(metadata)

    def assign_metadata(self, metadata: list[FileMetadata]) -> None:
        """
        Assigns locally available metadata to the file object.
        This is mostly used internally and can be ignored by most users.
        """
        for item in metadata:
            if item.path == self.path and item.is_latest:
                self._metadata = item
                return



"""
A basic executable interface to Coscine.

The program offers a set of actions that apply to the current selection
of Coscine project, resource and/or files.
Example:
    >>> coscine list
    This will list all projects of the Coscine user.

    >>> coscine list -p "My Coscine Project"
    This will list all resources inside the project "My Coscine Project".

    >>> coscine list -p "My Coscine Project" -r "My Coscine Resource"
    This will list all files inside the resource "My Coscine Resource"
    in the Coscine project "My Coscine Project".

    >>> coscine download -p "My Coscine Project"
    This will download the project "My Coscine Project".

    >>> coscine download -p "Project" -r "Resource" -f "File path"
    This will download the file "File path".
"""

import argparse
import coscine
import logging
from os import getenv
from pathlib import Path
from datetime import datetime

logging.basicConfig(level=logging.INFO)
argument_parser = argparse.ArgumentParser("coscine", description=__doc__, formatter_class=argparse.RawTextHelpFormatter)
# Positional arguments
ACTIONS = [
    "delete",
    "download",
    "list",
    "upload",
    "update"
]
argument_parser.add_argument("action", choices=ACTIONS)
# Optional arguments
argument_parser.add_argument("-v", "--version", action="version", version=coscine.ApiClient.BANNER)
argument_parser.add_argument("-e", "--endpoint", type=str, help="Set the Coscine API base url", default="https://coscine.rwth-aachen.de")
argument_parser.add_argument("-p", "--project", type=str, help="Coscine project display name")
argument_parser.add_argument("-r", "--resource", type=str, help="Coscine resource display name")
argument_parser.add_argument("-f", "--file", type=str, help="Coscine file path")
argument_parser.add_argument("-t", "--token", type=str, help="Coscine API token")
args = argument_parser.parse_args()
token = args.token if args.token else getenv("COSCINE_API_TOKEN")
if token is None:
    raise RuntimeError(
        "Specify a coscine API token via the command-line argument -t, "
        "--token or via the environment variable COSCINE_API_TOKEN!"
        " For more information visit https://docs.coscine.de/en/token/."
    )
client = coscine.ApiClient(token, base_url=args.endpoint)
if client.version != client.latest_version():
    logging.warning(
        f"Using version {client.version} of the Coscine Python SDK, but "
        f"a newer version {client.latest_version()} is available on PyPi! "
        "Consider updating the Coscine Python SDK via `pip install -U coscine` "
        "or `conda update coscine`."
    )
project  = client.project(args.project)     if args.project else None
resource = project.resource(args.resource)  if args.resource else None
if args.action not in ("upload", "update"):
    file = resource.file(args.file)         if args.file else None
else:
    file = str(args.file) if args.file else None


def action_delete():
    """
    Deletes Coscine projects or resources or files.
    """
    if not project:
        raise RuntimeError("Action 'delete' requires at least a project name!")
    elif not resource:
        if input(f"Delete Coscine project '{project.name}' (Y/N)? ").upper() == "Y":
            project.delete()
            print(f"Project {project.name} deleted!")
        else:
            print("Delete aborted!")
    elif not file:
        if input(f"Delete Coscine resource '{resource.name}' (Y/N)? ").upper() == "Y":
            resource.delete()
            print(f"Resource {resource.name} deleted!")
        else:
            print("Delete aborted!")
    else:
        if input(f"Delete Coscine file '{file.path}' (Y/N)? ").upper() == "Y":
            file.delete()
            print(f"File {file.path} deleted!")
        else:
            print("Delete aborted!")

def action_download():
    """
    Downloads Coscine projects or resources or files.
    """
    if not project:
        raise RuntimeError("Action 'download' requires at least a project name!")
    elif not resource:
        project.download()
    elif not file:
        resource.download()
    else:
        file.download()


def action_list():
    """
    Lists Coscine projects or resources or files.
    """
    if not project:
        for index, item in enumerate(client.projects()):
            print(f"Project [{index}]")
            print(f"  name: {item.name}")
            print(f"  display_name: {item.display_name}")
            print(f"  description: {(item.description[:75] + '...') if len(item.description) > 75 else item.description}")
    elif not resource:
        for index, item in enumerate(project.resources()):
            print(f"Resource [{index}]")
            print(f"  name: {item.name}")
            print(f"  display_name: {item.display_name}")
            print(f"  description: {(item.description[:75] + '...') if len(item.description) > 75 else item.description}")
    else:
        for index, item in enumerate(resource.files()):
            print(f"File [{index}] {item.path}")


def action_upload():
    """
    Uploads files to a Coscine resource. This will overwrite existing files.
    """
    if not (project and resource):
        raise RuntimeError("Action 'upload' requires a project and a resource name!")
    metadata = resource.metadata_form()
    print(
        "Because entering metadata from the command line is really annoying, "
        "required fields are automatically filled with garbage values. "
        "You can edit the metadata after the upload in the Coscine web interface "
        "or via a python script."
    )
    #print("To upload the file you must specify its metadata according to the resource metadata profile.")
    #print("You can leave non-required fields empty by just pressing Enter/Return.")
    #print("Required fields are marked with an asterisk, e.g. 'Field*' instead of 'Field'.")
    metadata.test()
    #for field in metadata.fields():
    #    field_name = field.name if not field.is_required else field.name + "*"
    #    if field.has_vocabulary:
    #        key = input(f"{field_name}: ")
    #        if key:
    #            value = field.vocabulary(key)
    #            metadata[field.name] = value
    #    elif field.has_selection:
    #        key = input(f"{field_name}: ")
    #        if key:
    #            value = field.selection(key)
    #            metadata[field.name] = value
    #    elif field.datatype == str:
    #        value = input(f"{field_name}: ")
    #        if value:
    #            metadata[field.name] = value
    #    elif field.datatype == int:
    #        value = input(f"{field_name}: ")
    #        if value:
    #            metadata[field.name] = int(value)
    #    elif field.datatype == float:
    #        value = input(f"{field_name}: ")
    #        if value:
    #            metadata[field.name] = float(value)
    #    elif field.datatype == datetime.date:
    #        value = input(f"{field_name}: ")
    #        if value:
    #            metadata[field.name] = datetime.strptime(value, r"%Y-%m-%d").date()
    with open(file, "rb") as fp:
        resource.upload(Path(file).name, fp, metadata, overwrite=True)


def action_update():
    with open(file, "rb") as fp:
        resource._upload_blob(Path(file).name, fp, use_put=True)


if args.action == "delete":
    action_delete()
if args.action == "download":
    action_download()
elif args.action == "list":
    action_list()
elif args.action == "upload":
    action_upload()
elif args.action == "update":
    action_update()



###############################################################################
# Coscine Python SDK
# Copyright (c) 2020-2025 RWTH Aachen University
# Licensed under the terms of the MIT License
# For more information on Coscine visit https://www.coscine.de/.
###############################################################################

"""
Coscine Python SDK package metadata
"""

__author__ = "RWTH Aachen University"
__copyright__ = "RWTH Aachen University"
__license__ = "MIT License"
__version__ = "0.11.12"



###############################################################################
# Coscine Python SDK
# Copyright (c) 2020-2025 RWTH Aachen University
# Licensed under the terms of the MIT License
# For more information on Coscine visit https://www.coscine.de/.
###############################################################################

"""
The Coscine Python SDK ships with its own set of exceptions.
All exceptions raised by the Coscine Python SDK are derived from
a common base exception class called "CoscineException".
"""


class CoscineException(Exception):
    """
    Coscine Python SDK base exception.
    Inherited by all other Coscine Python SDK exceptions.
    """


class AuthenticationError(CoscineException):
    """
    Failed to authenticate with the API token supplied by the user.
    """


class TooManyResults(CoscineException):
    """
    Two or more instances match the property provided by the user but
    the Coscine Python SDK expected just a single instance to match.
    """


class NotFoundError(CoscineException):
    """
    The droids you were looking for have not been found.
    Move along!
    """


class RequestRejected(CoscineException):
    """
    The request has reached the Coscine servers but has
    been rejected for whatever reason there may be. This
    exception is most likely thrown in case of ill-formatted
    requests.
    """



###############################################################################
# Coscine Python SDK
# Copyright (c) 2020-2025 RWTH Aachen University
# Licensed under the terms of the MIT License
# For more information on Coscine visit https://www.coscine.de/.
###############################################################################

import setuptools

PROJECT_URL = (
    "https://git.rwth-aachen.de/coscine/community-features/coscine-python-sdk"
)

README = ""

# Read README.md into $description to serve as the long_description on PyPi.
with open("README.md", "rt", encoding="utf-8") as fp:
    README = fp.read()


# Read package metadata into $about
about = {}
with open("src/coscine/__about__.py", "r", encoding="utf-8") as fp:
    exec(fp.read(), about)
__author__ = about["__author__"]
__version__ = about["__version__"]


setuptools.setup(
    name="coscine",
    version=__version__,
    description=(
        "The Coscine Python SDK provides a pythonic high-level "
        "interface to the Coscine REST API."
    ),
    long_description=README,
    long_description_content_type="text/markdown",
    author=__author__,
    author_email="coscine@itc.rwth-aachen.de",
    license="MIT License",
    packages=setuptools.find_packages(where="src"),
    keywords=["Coscine", "RWTH Aachen", "Research Data Management"],
    install_requires=[
        "boto3",
        "isodate",
        "pyshacl",
        "rdflib",
        "requests",
        "requests-cache",
        "requests-toolbelt",
        "tabulate",
        "tqdm"
    ],
    url=PROJECT_URL,
    project_urls={
        "Issues": PROJECT_URL + "/-/issues",
        "Documentation": (
            "https://coscine.pages.rwth-aachen.de/"
            "community-features/coscine-python-sdk/"
        )
    },
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Development Status :: 5 - Production/Stable",
        "Typing :: Typed"
    ],
    package_dir={"": "src"},
    python_requires=">=3.10"
)



###############################################################################
# Coscine Python SDK
# Copyright (c) 2020-2025 RWTH Aachen University
# Licensed under the terms of the MIT License
# For more information on Coscine visit https://www.coscine.de/.
###############################################################################

"""
Provides common classes shared among multiple modules.
"""

from datetime import date, datetime
from isodate import parse_datetime


class MaintenanceNotice:
    """
    Models maintenance notices set in Coscine.
    """

    _data: dict

    @property
    def title(self) -> str:
        """
        The title or name of the maintenance notice.
        """
        return self._data.get("displayName") or ""

    @property
    def link(self) -> str:
        """
        The URL link to the detailed maintenance notice.
        """
        return self._data.get("href") or ""

    @property
    def type(self) -> str:
        """
        The type of maintenance.
        """
        return self._data.get("type") or ""

    @property
    def body(self) -> str:
        """
        The body or description of the notice.
        """
        return self._data.get("body") or ""

    @property
    def starts_date(self) -> date:
        """
        Date when the maintenance goes active.
        """
        value: str = self._data.get("startsDate") or "1998-01-01T00:00:00Z"
        return parse_datetime(value).date()

    @property
    def ends_date(self) -> date:
        """
        Date when the maintenance ends.
        """
        value = self._data.get("endsDate") or "1998-01-01T00:00:00Z"
        return parse_datetime(value).date()

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.title


class SearchResult:
    """
    This class models the search results returned by Coscine upon
    a search request.
    """

    _data: dict

    @property
    def uri(self) -> str:
        """
        Link to the result (i.e. a project or resource or file).
        """
        return self._data.get("uri") or ""

    @property
    def type(self) -> str:
        """
        The search category the result falls into (e.g. project).
        """
        return self._data.get("type") or ""

    @property
    def source(self) -> str:
        """
        The source text that matches in some way or another the search query.
        """
        return self._data.get("source") or ""

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.uri


class ApiToken:
    """
    This class models the Coscine API token.
    """

    _data: dict

    @property
    def id(self) -> str:
        """
        Unique Coscine-internal identifier for the API token.
        """
        return self._data.get("id") or ""

    @property
    def name(self) -> str:
        """
        The name assigned to the API token by the creator upon creation.
        """
        return self._data.get("name") or ""

    @property
    def created(self) -> date:
        """
        Timestamp of when the API token was created.
        """
        value = self._data.get("creationDate") or "1998-01-01T00:00:00Z"
        return parse_datetime(value).date()

    @property
    def expires(self) -> date:
        """
        Timestamp of when the API token will expire.
        """
        value = self._data.get("expiryDate") or "1998-01-01T00:00:00Z"
        return parse_datetime(value).date()

    @property
    def expired(self) -> bool:
        """
        Evaluates to True if the API token is expired.
        """
        return datetime.now().date() > self.expires

    @property
    def owner(self) -> str:
        """
        Unique Coscine-internal user id of the owner of the API token.
        """
        return self._data["owner"]["id"]

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.name


class Language:
    """
    Models the languages available in Coscine.
    """

    _data: dict

    @property
    def id(self) -> str:
        """
        Unique and constant Coscine internal identifier for
        the respective language option.
        """
        return self._data.get("id") or ""

    @property
    def name(self) -> str:
        """
        The full name of the language option.
        """
        return self._data.get("displayName") or ""

    @property
    def abbreviation(self) -> str:
        """
        The abbreviated name of the language option.
        """
        return self._data.get("abbreviation") or ""

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.name


class AcademicTitle:
    """
    Models the Academic Titles available in Coscine.
    """

    _data: dict

    @property
    def id(self) -> str:
        """
        Unique and constant Coscine internal identifier for
        the respective Academic Title.
        """
        return self._data.get("id") or ""

    @property
    def name(self) -> str:
        """
        The name of the Academic Title, e.g. "Prof." or "Dr."
        """
        return self._data.get("displayName") or ""

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.name


class Discipline:
    """
    Models the disciplines available in Coscine.
    """

    _data: dict

    @property
    def id(self) -> str:
        """
        The Coscine-internal unique identifier for the discipline.
        """
        return self._data.get("id") or ""

    @property
    def uri(self) -> str:
        """
        The uri of the discipline.
        """
        return self._data.get("uri") or ""

    @property
    def name(self) -> str:
        """
        The human-readable name of the discipline.
        """
        return self._data.get("displayNameEn") or ""

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.name

    def serialize(self) -> dict:
        """
        Returns the machine-readable representation of the discipline
        data instance.
        """
        return self._data


class License:
    """
    Models the licenses available in Coscine.
    """

    _data: dict

    @property
    def id(self) -> str:
        """
        The Coscine-internal unique identifier for the license.
        """
        return self._data.get("id") or ""

    @property
    def name(self) -> str:
        """
        The human-readable name of the license.
        """
        return self._data.get("displayName") or ""

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.name

    def serialize(self) -> dict:
        """
        Returns the machine-readable representation of the license
        data instance.
        """
        return self._data


class Organization:
    """
    Models organization information for organizations in Coscine.
    """

    _data: dict

    @property
    def uri(self) -> str:
        """
        The organization's ror uri.
        """
        return self._data.get("uri") or ""

    @property
    def name(self) -> str:
        """
        The full name of the organization.
        """
        return self._data.get("displayName") or ""

    @property
    def email(self) -> str:
        """
        Contact email address of the organization.
        """
        return self._data.get("email") or ""

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.name

    def serialize(self) -> dict:
        """
        Returns the machine-readable representation of the organization
        data instance.
        """
        return self._data


class User:
    """
    This class provides an interface around userdata in Coscine.
    """

    _data: dict

    @property
    def id(self) -> str:
        """
        The unique Coscine-internal user id for a user.
        """
        return self._data.get("id") or ""

    @property
    def display_name(self) -> str:
        """
        The full name of a Coscine user as displayed
        in the Coscine web interface.
        """
        return self._data.get("displayName") or ""

    @property
    def first_name(self) -> str:
        """
        The first name of a Coscine user.
        """
        return self._data.get("givenName") or ""

    @property
    def last_name(self) -> str:
        """
        The family name of a Coscine user.
        """
        return self._data.get("familyName") or ""

    @property
    def email(self) -> str | list[str]:
        """
        The email address or list of email addresses of a user.
        In case the user has not associated an email address with their
        account 'None' is returned.
        """
        if "email" in self._data:
            return self._data.get("email") or ""
        return self._data.get("emails") or []

    @property
    def title(self) -> AcademicTitle | None:
        """
        The academic title of a user.
        In case the user has not set a title in their user profile
        'None' is returned.
        """
        title = self._data.get("title")
        if title is not None:
            title = AcademicTitle(title)
        return title

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.display_name


class Visibility:
    """
    Models the visibility settings available in Coscine.
    """

    _data: dict

    @property
    def id(self) -> str:
        """
        Coscine-internal identifier for the visibility setting.
        """
        return self._data.get("id") or ""

    @property
    def name(self) -> str:
        """
        Human-readable name of the visibility setting.
        """
        return self._data.get("displayName") or ""

    def __init__(self, data: dict) -> None:
        self._data = data

    def __str__(self) -> str:
        return self.name

    def serialize(self) -> dict:
        """
        Returns the machine-readable representation of the visibility
        data instance.
        """
        return self._data
